try:
    import IPython
    shell = IPython.get_ipython()
    shell.enable_matplotlib(gui='qt')
except:
    pass


from MJOLNIR import _tools # Useful tools useful across MJOLNIR
try:
    import _tools as _guitools
    from Views import BraggListManager
except ImportError:
    import MJOLNIRGui.src.main.python._tools as _guitools
    from MJOLNIRGui.src.main.python.Views import BraggListManager
import numpy as np
import matplotlib.pyplot as plt
import sys
import datetime
from time import sleep

from os import path
import os
version = '0.9.9'
plt.ion()
from PyQt5 import QtWidgets, QtCore, QtGui, Qt
try:
    #from MJOLNIR_GUI_ui import Ui_MainWindow  
    
    from Views.main import Ui_MainWindow
    from Views.DataSetManager import DataSetManager
    from Views.View3DManager import View3DManager
    from Views.QELineManager import QELineManager
    from Views.QPlaneManager import QPlaneManager
    from Views.Cut1DManager import Cut1DManager
    from Views.MaskManager import MaskManager
    from Views.Raw1DManager import Raw1DManager
    from Views.NormalizationManager import NormalizationManager
    from Views.MolecularCalculationManager import MolecularCalculationManager
    from Views.PredictionToolManager import PredictionToolManager
    from Views.CalculatorManager import CalculatorManager
    from Views.SubtractionManager import SubtractionManager
    from Views.collapsibleBox import CollapsibleBox
    from Views.ElectronicLogBookManager import ElectronicLogBookManager
    from MJOLNIR_Data import GuiDataFile,GuiDataSet
    from DataModels import DataSetModel,DataFileModel
    from StateMachine import StateMachine
    from GuiStates import empty,partial,raw,converted
    from AboutDialog import AboutDialog
    from HelpDialog import HelpDialog
    from generateScripts import initGenerateScript,setupGenerateScript
    from _tools import loadSetting,updateSetting,ProgressBarDecoratorArguments
    
except ModuleNotFoundError:
    sys.path.append('.')
    
    #from .MJOLNIR_GUI_ui import Ui_MainWindow  
    from MJOLNIRGui.src.main.python.Views.main import Ui_MainWindow
    from MJOLNIRGui.src.main.python.Views.DataSetManager import DataSetManager
    from MJOLNIRGui.src.main.python.Views.View3DManager import View3DManager
    from MJOLNIRGui.src.main.python.Views.QELineManager import QELineManager
    from MJOLNIRGui.src.main.python.Views.QPlaneManager import QPlaneManager
    from MJOLNIRGui.src.main.python.Views.Cut1DManager import Cut1DManager
    from MJOLNIRGui.src.main.python.Views.MaskManager import MaskManager
    from MJOLNIRGui.src.main.python.Views.Raw1DManager import Raw1DManager
    from MJOLNIRGui.src.main.python.Views.NormalizationManager import NormalizationManager
    from MJOLNIRGui.src.main.python.Views.MolecularCalculationManager import MolecularCalculationManager
    from MJOLNIRGui.src.main.python.Views.PredictionToolManager import PredictionToolManager
    from MJOLNIRGui.src.main.python.Views.CalculatorManager import CalculatorManager
    from MJOLNIRGui.src.main.python.Views.SubtractionManager import SubtractionManager
    from MJOLNIRGui.src.main.python.Views.collapsibleBox import CollapsibleBox
    from MJOLNIRGui.src.main.python.Views.ElectronicLogBookManager import ElectronicLogBookManager
    from MJOLNIRGui.src.main.python.MJOLNIR_Data import GuiDataFile,GuiDataSet
    from MJOLNIRGui.src.main.python.DataModels import DataSetModel,DataFileModel
    from MJOLNIRGui.src.main.python.StateMachine import StateMachine
    from MJOLNIRGui.src.main.python.GuiStates import empty,partial,raw,converted
    from MJOLNIRGui.src.main.python.AboutDialog import AboutDialog
    from MJOLNIRGui.src.main.python.HelpDialog import HelpDialog
    from MJOLNIRGui.src.main.python.generateScripts import initGenerateScript,setupGenerateScript
    from MJOLNIRGui.src.main.python._tools import loadSetting,updateSetting,ProgressBarDecoratorArguments

import sys


from pathlib import Path
home = str(Path.home())


####

# Naming convention: WhereInGui_description_type
# Append _function if it is a function
# E.g.: View3D_plot_button and View3D_plot_button_function

#Headlines so far are:
#DataSet, View3D, QE Cut, QPlane, Cut1D, Raw1D


class MJOLNIRMainWindow(QtWidgets.QMainWindow):
    mask_changed = QtCore.pyqtSignal()
    state_changed = QtCore.pyqtSignal(str,str,name='stateChanged')
    def __init__(self,AppContext):

        super(MJOLNIRMainWindow, self).__init__()

        self.ui = Ui_MainWindow()
        self.AppContext = AppContext
        self.version = version
        ### Settings saved in .MJOLNIRGuiSettings
        self.settingsFile = path.join(home,'.MJOLNIRGuiSettings')
        self.views = []
        guiSettings = loadSetting(self.settingsFile,'guiSettings')
        self.colormap = 'viridis'
        self.currentFolder = ''

        self.isClosing = False 
        self.ui.setupUi(self)
        self.update()

        icon = QtGui.QIcon()
        icon.addPixmap(QtGui.QPixmap(self.AppContext.get_resource('Icons/Own/MJOLNIR.png')))
        self.setWindowIcon(icon)

        # List to hold all views that need to be setup
        self.views = []
        ## Set up DataSetManager
        self.ui.dataSetManager = DataSetManager(None,self)
        vlay = QtWidgets.QVBoxLayout(self.ui.fixedOpen)
        vlay.addWidget(self.ui.dataSetManager)

        vlay.addStretch()

        self.update()
        self.views.append(self.ui.dataSetManager)

        self.windows = [] # Holder for generated plotting windows
        self.figureList = [] # Holder for selectable window list

        # Lists of views in shown order
        self.nameList = ['View3D','QE cuts','Q plane','1D cuts','1D raw data','Masking'] # 'Normalization'
        self.viewClasses = [View3DManager,QELineManager,QPlaneManager,Cut1DManager,Raw1DManager]#[View3D,View3D,View3D,Cut1D,Raw1D] # NormalizationManager
        self.startState = [True,False,False,False,True,False] # If not collapsed #False

        # Find correct layout to insert views
        vlay = QtWidgets.QVBoxLayout(self.ui.collapsibleContainer)
        # Insert all views
        self.boxContainers = []
        for name,Type,state in zip(self.nameList,self.viewClasses,self.startState):
            self.update()
            box = CollapsibleBox(name,startState=state)
            self.boxContainers.append(box)
            vlay.addWidget(box)
            lay = QtWidgets.QVBoxLayout()

            widget = Type(guiWindow=self)
            #if Type == NormalizationManager: # Get a reference to the sample manager directly in self
            #    self.normalizationManager = widget
            self.views.append(widget)
            lay.addWidget(widget)
           
            box.setContentLayout(lay)
        
        vlay.setAlignment(QtCore.Qt.AlignTop)

        self.dataSets = []

        self.current_timer = None
        
        self.blockItems = [getattr(self.ui,item) for item in self.ui.__dict__ if '_button' in item[-7:]] # Collect all items to block on calls
        self.blockItems.append(self.ui.DataSet_binning_comboBox)

        self.lineEdits = [getattr(self.ui,item) for item in self.ui.__dict__ if '_lineEdit' in item[-9:]] # Collect all lineedits
        self.radioButtons = [getattr(self.ui,item) for item in self.ui.__dict__ if '_radioButton' in item] # Collect all radiobuttons
        self.spinBoxes = [getattr(self.ui,item) for item in self.ui.__dict__ if '_spinBox' in item[-8:]] # Collect all spinboxes
        self.checkBoxes = [getattr(self.ui,item) for item in self.ui.__dict__ if '_checkBox' in item[-9:]] # Collect all checkboxes


        self.ui.actionSave_GUI_state.setShortcut("Ctrl+S")
        self.ui.actionLoad_GUI_state.setShortcut("Ctrl+O")
        self.ui.actionExit.setShortcut("Ctrl+Q")

        self.ui.actionHelp.setShortcut("Ctrl+?")

        self.log = _guitools.log(self.ui.textBrowser,button=self.ui.log_reset_btn)
        self.ui.log_reset_btn.clicked.connect(self.clearStatus)
        self._clearStatus() # Call underscore method to avoid decorator
        

        self.ui.progressBar._text = 'Ready'
        self.ui.progressBar.setAlignment(QtCore.Qt.AlignCenter)
        

        def setText(self, text):
            self._text = text

        def text(self):
            if not self.value == 0:
                return self._text + ' ' + self._oldText()
            else:
                return self._text

        self.ui.progressBar.setText = lambda text: setText(self.ui.progressBar,text)
        self.ui.progressBar._oldText = self.ui.progressBar.text
        self.ui.progressBar.text = lambda: text(self.ui.progressBar)

        self.braggPoints = None

        self.update()
        initGenerateScript(self)

        for view in self.views: # Run through all views to set them up
            view.setup()

        self.maskingManager = MaskManager(self)

        setupGenerateScript(self)
        self.update()
        self.setupMenu()
        self.update()
        self.setupStateMachine()
        self.update()
        self.stateMachine.run()
        self.update()
        self.loadFolder() # Load last folder as default 
        self.loadedGuiSettings = None
        self.ui.menubar.setNativeMenuBar(False)
        
        if sys.platform.lower() == 'darwin':
        ## Update image of arrows to correct style on mac
            correctedArrows = """QToolButton::down-arrow {
        image: url("""+self.AppContext.get_resource('down.png')+""");
    }

    QToolButton::right-arrow {
        image: url("""+self.AppContext.get_resource('right.png')+""");
    }"""

            self.setStyleSheet(self.styleSheet()+correctedArrows)

        # Set up preset for electronic logbook
        self.logbookPreset = ['A3','twoTheta','sampleName','title','scanCommand','ei']

    def setupMenu(self): # Set up all QActions and menus
        self.ui.actionExit.setIcon(QtGui.QIcon(self.AppContext.get_resource('Icons/Own/cross-button.png')))
        self.ui.actionExit.setToolTip('Exit the application') 
        self.ui.actionExit.setStatusTip(self.ui.actionExit.toolTip())
        self.ui.actionExit.triggered.connect(self.close)

        self.ui.actionAbout.setIcon(QtGui.QIcon(self.AppContext.get_resource('Icons/Own/information-button.png')))
        self.ui.actionAbout.setToolTip('Show About') 
        self.ui.actionAbout.setStatusTip(self.ui.actionAbout.toolTip())
        self.ui.actionAbout.triggered.connect(self.about)

        self.ui.actionHelp.setIcon(QtGui.QIcon(self.AppContext.get_resource('Icons/Own/question-button.png')))
        self.ui.actionHelp.setToolTip('Show Help') 
        self.ui.actionHelp.setStatusTip(self.ui.actionHelp.toolTip())
        self.ui.actionHelp.triggered.connect(self.help)

        self.ui.actionSubtraction.setIcon(QtGui.QIcon(self.AppContext.get_resource('Icons/Own/question-button.png')))
        self.ui.actionSubtraction.setToolTip('Show Subtraction Help') 
        self.ui.actionSubtraction.setStatusTip(self.ui.actionSubtraction.toolTip())
        self.ui.actionSubtraction.triggered.connect(self.subtractionHelp)

        self.ui.actionSave_GUI_state.setIcon(QtGui.QIcon(self.AppContext.get_resource('Icons/Own/folder-save.png')))
        self.ui.actionSave_GUI_state.setToolTip('Save current Gui setup') 
        self.ui.actionSave_GUI_state.setStatusTip(self.ui.actionSave_GUI_state.toolTip())
        self.ui.actionSave_GUI_state.triggered.connect(self.saveCurrentGui)
        #self.actionSave_GUI_state_shortcut = QtWidgets.QShortcut(QtGui.QKeySequence("Ctrl+S"), self)
        #self.ui.actionSave_GUI_state.triggered.connect(self.saveCurrentGui)

        self.ui.actionLoad_GUI_state.setIcon(QtGui.QIcon(self.AppContext.get_resource('Icons/Own/folder--arrow.png')))
        self.ui.actionLoad_GUI_state.setToolTip('Load Gui setup') 
        self.ui.actionLoad_GUI_state.setStatusTip(self.ui.actionLoad_GUI_state.toolTip())
        self.ui.actionLoad_GUI_state.triggered.connect(self.loadGui)
        #self.actionLoad_GUI_state_shortcut = QtWidgets.QShortcut(QtGui.QKeySequence("Ctrl+O"), self)
        #self.ui.actionLoad_GUI_state.triggered.connect(self.loadGui)

        self.ui.actionGenerate_View3d_script.setIcon(QtGui.QIcon(self.AppContext.get_resource('Icons/Own/script-3D.png')))
        self.ui.actionGenerate_View3d_script.setToolTip('Generate 3D Script') 
        self.ui.actionGenerate_View3d_script.setStatusTip(self.ui.actionGenerate_View3d_script.toolTip())
        self.ui.actionGenerate_View3d_script.triggered.connect(self.generate3DScript)

        self.ui.actionGenerate_QELine_script.setIcon(QtGui.QIcon(self.AppContext.get_resource('Icons/Own/script-QE.png')))
        self.ui.actionGenerate_QELine_script.setToolTip('Generate QE Cut Script') 
        self.ui.actionGenerate_QELine_script.setStatusTip(self.ui.actionGenerate_QELine_script.toolTip())
        self.ui.actionGenerate_QELine_script.triggered.connect(self.generateQEScript)
        
        self.ui.actionGenerate_QPlane_script.setIcon(QtGui.QIcon(self.AppContext.get_resource('Icons/Own/script-QP.png')))
        self.ui.actionGenerate_QPlane_script.setToolTip('Generate QPlane Script') 
        self.ui.actionGenerate_QPlane_script.setStatusTip(self.ui.actionGenerate_QPlane_script.toolTip())
        self.ui.actionGenerate_QPlane_script.triggered.connect(self.generateQPlaneScript)
        
        self.ui.actionGenerate_1d_script.setIcon(QtGui.QIcon(self.AppContext.get_resource('Icons/Own/script-1D.png')))
        self.ui.actionGenerate_1d_script.setToolTip('Generate Cut1D Script') 
        self.ui.actionGenerate_1d_script.setStatusTip(self.ui.actionGenerate_1d_script.toolTip())
        self.ui.actionGenerate_1d_script.triggered.connect(self.generateCut1DScript)
        
        self.ui.action_masking_gui.setIcon(QtGui.QIcon(self.AppContext.get_resource('Icons/Own/mask-open.png')))
        self.ui.action_masking_gui.setToolTip('Masking Gui') 
        self.ui.action_masking_gui.setStatusTip(self.ui.action_masking_gui.toolTip())
        self.ui.action_masking_gui.triggered.connect(self.maskingManager.setWindowVisible)
        self.ui.action_masking_gui.setShortcut("Ctrl+M")


        self.ui.actionSettings.setIcon(QtGui.QIcon(self.AppContext.get_resource('Icons/Own/settings.png')))
        self.ui.actionSettings.setDisabled(False)
        self.ui.actionSettings.setToolTip('Change View Settings') 
        self.ui.actionSettings.setStatusTip(self.ui.actionSettings.toolTip())
        self.ui.actionSettings.triggered.connect(self.settingsDialog)

        
        self.ui.actionClose_Windows.setIcon(QtGui.QIcon(self.AppContext.get_resource('Icons/Own/CloseWindows.png')))
        self.ui.actionClose_Windows.setDisabled(False)
        self.ui.actionClose_Windows.setToolTip('Close All Plotting Windows') 
        self.ui.actionClose_Windows.setStatusTip(self.ui.actionClose_Windows.toolTip())
        self.ui.actionClose_Windows.triggered.connect(self.closeWindows)


        self.ui.actionGenerate_Normalization.setIcon(QtGui.QIcon(self.AppContext.get_resource('Icons/Own/ruler.png')))
        self.ui.actionGenerate_Normalization.setDisabled(False)
        self.ui.actionGenerate_Normalization.setToolTip('Generate a script to normalize data absolutely') 
        self.ui.actionGenerate_Normalization.setStatusTip(self.ui.actionGenerate_Normalization.toolTip())
        self.ui.actionGenerate_Normalization.triggered.connect(self.absolutNormalizationTool)

        self.ui.actionPrediction_Tool.setIcon(QtGui.QIcon(self.AppContext.get_resource('Icons/Own/predict.png')))
        self.ui.actionPrediction_Tool.setDisabled(False)
        self.ui.actionPrediction_Tool.setToolTip('Predict scan coverage') 
        self.ui.actionPrediction_Tool.setStatusTip(self.ui.actionPrediction_Tool.toolTip())
        self.ui.actionPrediction_Tool.triggered.connect(self.predictionTool)
        self.ui.actionPrediction_Tool.setShortcut("Ctrl+P")

        self.ui.actionCalculate_Molecular_Weight.setIcon(QtGui.QIcon(self.AppContext.get_resource('Icons/Own/balance.png')))
        self.ui.actionCalculate_Molecular_Weight.setDisabled(False)
        self.ui.actionCalculate_Molecular_Weight.setToolTip('Calculate Molecular Mass from Chemical Formula') 
        self.ui.actionCalculate_Molecular_Weight.setStatusTip(self.ui.actionCalculate_Molecular_Weight.toolTip())
        self.ui.actionCalculate_Molecular_Weight.triggered.connect(self.molarMassTool)

        self.ui.actionNeutron_Calculations.setIcon(QtGui.QIcon(self.AppContext.get_resource('Icons/Own/calculator.png')))
        self.ui.actionNeutron_Calculations.setDisabled(False)
        self.ui.actionNeutron_Calculations.setToolTip('Calculate standard neutron quantities') 
        self.ui.actionNeutron_Calculations.setStatusTip(self.ui.actionNeutron_Calculations.toolTip())
        self.ui.actionNeutron_Calculations.triggered.connect(self.neutronCalculationTool)

        self.ui.actionElectronic_Logbook.setIcon(QtGui.QIcon(self.AppContext.get_resource('Icons/Own/book--pencil.png')))
        self.ui.actionElectronic_Logbook.setToolTip('Generate Electronic Logbook from files') 
        self.ui.actionElectronic_Logbook.setStatusTip(self.ui.actionElectronic_Logbook.toolTip())
        self.ui.actionElectronic_Logbook.triggered.connect(self.electronicLogbookTool)
        self.ui.actionElectronic_Logbook.setShortcut("Ctrl+E")

        self.ui.actionSubtraction_Of_DataSets.setIcon(QtGui.QIcon(self.AppContext.get_resource('Icons/Own/subtract.png')))
        self.ui.actionSubtraction_Of_DataSets.setToolTip('Perform Subtraction of two DataSets') 
        self.ui.actionSubtraction_Of_DataSets.setStatusTip(self.ui.actionSubtraction_Of_DataSets.toolTip())
        self.ui.actionSubtraction_Of_DataSets.triggered.connect(self.subtractionManager)
        self.ui.actionSubtraction_Of_DataSets.setShortcut("Ctrl+D")

        self.ui.View3D_CurratAxe_button.clicked.connect(self.openBraggListWindow)

    def getProgressBarValue(self):
        return self.ui.progressBar.value

    def addProgressBarValue(self,value):
        currentValue = self.getProgressBarValue()+value
        self.ui.progressBar.setValue(currentValue)
        self.ui.progressBar.value = currentValue
        

    def setProgressBarValue(self,value):
        if not hasattr(self,'ui.progressBar.value'):
            self.ui.progressBar.value = 0
        
        self.ui.progressBar.setValue(value)
        self.ui.progressBar.value = value

    def setProgressBarLabelText(self,text):
        if self.current_timer:
            self.current_timer.stop()
        #self.ui.progressBar_label.setText(text)
        self.ui.progressBar.setText(text)

    def setProgressBarMaximum(self,value):
        self.ui.progressBar.setMaximum(value)

    def resetProgressBar(self):
        self.setProgressBarValue(0)
        self.setProgressBarLabelText('Ready')

    def saveSettingsDialog(self,event):
        res = QtWidgets.QMessageBox.question(self,
                                    "Exit - Save Gui Settings",
                                    "Do you want to save Gui Settings?",
                                    QtWidgets.QMessageBox.Save | QtWidgets.QMessageBox.No | QtWidgets.QMessageBox.Cancel)
        
        if res == QtWidgets.QMessageBox.Save:
            if self.saveCurrentGui(): # successful saving
                self.closeWindows()
                event.accept()
            else:
                event.ignore()
                return False
        elif res == QtWidgets.QMessageBox.No:
            self.closeWindows()
            event.accept()
            return 1
        else:
            event.ignore()
            return 0

    def quitDialog(self,event):
        res = QtWidgets.QMessageBox.question(self,
                                    "Exit",
                                    "Do you want to exit the Gui?",
                                    QtWidgets.QMessageBox.Yes | QtWidgets.QMessageBox.No)
        
        if res == QtWidgets.QMessageBox.Yes:
            self.closeWindows()
            event.accept()
            return 1
        else:
            event.ignore()
            return 0

    def restart(self):
        QtCore.QCoreApplication.quit()
        status = QtCore.QProcess.startDetached(sys.executable, sys.argv)
        print(status)

    def closeEvent(self, event):
        if self.loadedGuiSettings is None:
            if not self.saveSettingsDialog(event): # The dialog is cancelled
                return
        
        elif np.all([s1==s2 for s1,s2 in zip(self.loadedGuiSettings.values(),self.generateCurrentGuiSettings().values())]):
            if not self.quitDialog(event):
                return

        else:
            if not self.saveSettingsDialog(event): # The dialog is cancelled
                return
        self.isClosing = True
        self.maskingManager.close()
        self.closeWindows()

    @ProgressBarDecoratorArguments(runningText='Closing Windows',completedText='Windows Closed')
    def closeWindows(self):
        if hasattr(self,'windows'):
            for window in self.windows:
                try:
                    plt.close(window)
                except:
                    try:
                        window.close()
                    except:
                        pass
            for figureList in self.figureList:
                figureList.closeAll()
        return True

    def about(self):
        dialog = AboutDialog(self.AppContext.get_resource('About.txt'),version=self.version,icon=QtGui.QIcon(self.AppContext.get_resource('Icons/Own/information-button.png')))
        dialog.exec_()

    def help(self):
        dialog = HelpDialog(self.AppContext.get_resource('Help.txt'),guiWindow = self)
        dialog.exec_()

    def subtractionHelp(self):
        dialog = HelpDialog(self.AppContext.get_resource('SubtractionHelp.txt'),guiWindow = self)
        dialog.exec_()


    def setupStateMachine(self):
        self.stateMachine = StateMachine([empty,partial,raw,converted],self)

    def update(self):
        QtWidgets.QApplication.processEvents()
        QtWidgets.QApplication.processEvents()
        

    @ProgressBarDecoratorArguments(runningText='Saving Gui Settings',completedText='Gui Settings Saved',failedText='Cancelled')
    def saveCurrentGui(self): # save data set and files in format DataSetNAME DataFileLocation DataFileLocation:DataSetNAME
        #DataSet = [self.dataSets[I].name for I in range(self.DataSetModel.rowCount(None))]
        
        settingsDict = self.generateCurrentGuiSettings(updateProgressBar=True)
        if not hasattr(self,'loadedSettingsFile'):
            self.loadedSettingsFile = home
        saveSettings,_ = QtWidgets.QFileDialog.getSaveFileName(self, 'Save File',self.loadedSettingsFile)

        if saveSettings is None or saveSettings == '':
            return False

        if not saveSettings.split('.')[-1] == 'MJOLNIRGuiSettings':
            saveSettings+='.MJOLNIRGuiSettings'

        if os.path.exists(saveSettings): # if file already exists user has been asked to delete it
            os.remove(saveSettings)
        for key,value in settingsDict.items():
            updateSetting(saveSettings,key,value)
        self.loadedGuiSettings = self.generateCurrentGuiSettings()
        return True


    def generateCurrentGuiSettings(self,updateProgressBar=False):
        saveString = []
        
        if updateProgressBar: self.setProgressBarMaximum(len(self.DataSetModel.dataSets))

        for i,ds in enumerate(self.DataSetModel.dataSets):
            dsDict = {'name':ds.name}
            
            localstring = [df.fileLocation if df.type != 'nxs' else df.original_file.fileLocation for df in ds]
            dsDict['files']=localstring
            dsDict['binning'] = [None if df.type != 'nxs' else df.binning for df in ds]
            if not ds.background is None:
                dsDict['background'] = ds.background
                dsDict['convertBeforeSubtract'] = ds.convertBeforeSubtract
            if not ds._maskingObject is None:
                dsDict['maskingObject'] = ds._maskingObject
            saveString.append(dsDict)
            if updateProgressBar: self.setProgressBarValue((i+1))

        lineEditString = self.generateCurrentLineEditSettings()
        radioButtonString = self.generateCurrentRadioButtonSettings()
        spinBoxString = self.generateCurrentSpinBoxSettings()
        checkBoxString = self.generateCurrentcheckBoxSettings()
        fileDir = ''
        logbookPreset = self.logbookPreset.copy()
        if hasattr(self,'predictionSettings'):
            predictionSettings  = self.predictionSettings
        else:
            predictionSettings = {}

        infos = self.DataFileInfoModel.currentInfos()
        guiSettings = self.guiSettings()
        returnDict = {'dataSet':saveString, 'lineEdits':lineEditString, 'radioButtons': radioButtonString,'spinBoxes':spinBoxString,
                      'checkBoxes':checkBoxString,'fileDir':fileDir, 'infos':infos, 'guiSettings':guiSettings,
                      'predictionSettings':predictionSettings,'logbookPreset':logbookPreset}
        return returnDict

    def generateCurrentLineEditSettings(self):
        lineEditValueString = {}
        for item in self.lineEdits:
            lineEditValueString[item.objectName()] = item.text()
        return lineEditValueString

    def generateCurrentSpinBoxSettings(self):
        spinBoxValueString = {}
        for item in self.spinBoxes:
            spinBoxValueString[item.objectName()] = item.value()
        return spinBoxValueString

    def generateCurrentcheckBoxSettings(self):
        chechValueString = {}
        for item in self.checkBoxes:
            chechValueString[item.objectName()] = item.isChecked()
        return chechValueString

    def generateCurrentRadioButtonSettings(self):
        radioButtonString = {}
        for item in self.radioButtons:
            radioButtonString[item.objectName()] = item.isChecked()
        return radioButtonString

    def loadFolder(self):
        fileDir = loadSetting(self.settingsFile,'fileDir')
        if not fileDir is None:
            self.setCurrentDirectory(fileDir)


    @ProgressBarDecoratorArguments(runningText='Loading gui settings',completedText='Loading Done',failedText='Cancelled')
    def loadGui(self,presetFileLocation=None):
        
        # Load saveFile
        if not hasattr(self,'loadedSettingsFolder'):
            folder = home
        else:
            folder = self.loadedSettingsFolder
        
        if presetFileLocation is None: # When no file is provided, open file dialogue
            settingsFile,_ = QtWidgets.QFileDialog.getOpenFileName(self,"Open GUI settings file", folder,"Setting (*.MJOLNIRGuiSettings);;All Files (*)")
        else:
            settingsFile = presetFileLocation
        self.update()
        self.loadedSettingsFolder = os.path.dirname(settingsFile)
        self.loadedSettingsFile = settingsFile
        
        if settingsFile is None or settingsFile == '':
            return False
        
        self.setProgressBarLabelText('Deleating Old Data Sets and Files')
        while self.DataSetModel.rowCount(None)>0:
            self.DataSetModel.delete(self.DataSetModel.getCurrentDatasetIndex())
        else:
            self.DataSetModel.layoutChanged.emit()
            self.DataFileModel.updateCurrentDataSetIndex()
        self.update()
        dataSetString = loadSetting(settingsFile,'dataSet')
        if not dataSetString is None:
            totalFiles = np.sum([len(dsDict['files'])+np.sum(1-np.array([d is None for d in dsDict['binning']]))+1 for dsDict in dataSetString])+1
            # Get estimate of total number of data files
            self.setProgressBarMaximum(totalFiles)
            
            self.setProgressBarValue(0)
            logbookPreset = loadSetting(settingsFile,'logbookPreset')
            if not logbookPreset is None:
                self.logbookPreset = logbookPreset
            

            for dsDict in dataSetString:
                DSName = dsDict['name']
                self.setProgressBarLabelText('Loading Data Set \''+DSName+'\'')   
                files = dsDict['files'] # data or foreground data
                if 'background' in dsDict: # Data set is a subtracted set!
                    background = dsDict['background']
                    if len(files)!=0: # If files in dataset, continue
                        dfs = []
                        for dfLocation in files:
                            df = GuiDataFile(dfLocation)
                            self.update()
                            dfs.append(df)
                            self.addProgressBarValue(0.5)
                        dfsBG = []
                        for dfLocation in background:
                            df = GuiDataFile(dfLocation)
                            self.update()
                            dfsBG.append(df)
                            self.addProgressBarValue(0.5)
                        
                        
                        foreground_ds = GuiDataSet(name='fg',dataFiles=dfs)
                        background_ds = GuiDataSet(name='bg',dataFiles=dfsBG)
                        binnings = dsDict['binning']
                        if dsDict['convertBeforeSubtract']:
                            self.setProgressBarLabelText('Converting Data Set \''+DSName+'\'')   
                            for df,binning in zip(foreground_ds,binnings): # Give the correct binning
                                df.binning = binning
                            foreground_ds.convertDataFile(guiWindow=self,setProgressBarMaximum=False,progressUpdate=0.5,printFunction=self.writeToStatus)
                            self.update()
                            for df,binning in zip(background_ds,binnings): # Assume binning is the same across data sets
                                df.binning = binning
                            background_ds.convertDataFile(guiWindow=self,setProgressBarMaximum=False,progressUpdate=0.5,printFunction=self.writeToStatus)
                            self.update()
                            

                        temp = foreground_ds-background_ds
                        
                        ds = GuiDataSet(name=DSName,dataSet=temp)
                        
                        ds.background = background
                        ds.convertBeforeSubtract = dsDict['convertBeforeSubtract']
                        
                        if not ds.convertBeforeSubtract: # Convert after subtraction if needed
                            if not np.any([b is None for b in dsDict['binning']]):
                                self.setProgressBarLabelText('Converting Data Set \''+DSName+'\'')   
                                ds.convertDataFile(guiWindow=self,setProgressBarMaximum=False,printFunction=self.writeToStatus)
                                self.update()
                        if 'maskingObject' in dsDict:
                            maskingObject = dsDict['maskingObject']                        
                            self.setProgressBarLabelText('Applying Mask to \''+DSName+'\'') 
                            ds.mask = maskingObject(ds)
                            ds._maskingObject = maskingObject

                else: # Regular dataset
                        
                    dfs = None
                    if len(files)!=0: # If files in dataset, continue
                        dfs = []
                        for dfLocation in files:
                            df = GuiDataFile(dfLocation)
                            self.update()
                            dfs.append(df)
                            self.addProgressBarValue(1)
                    ds = GuiDataSet(name=DSName,dataFiles=dfs)
                    if 'binning' in dsDict:
                        if not np.any([b is None for b in dsDict['binning']]):
                            binnings = dsDict['binning']
                            for df,binning in zip(ds,binnings):
                                df.binning = binning
                            self.setProgressBarLabelText('Converting Data Set \''+DSName+'\'')     
                            ds.convertDataFile(guiWindow=self,setProgressBarMaximum=False,printFunction=self.writeToStatus)
                            self.update()

                    if 'maskingObject' in dsDict:
                        maskingObject = dsDict['maskingObject']                        
                        ds.mask = maskingObject(ds)
                        ds._maskingObject = maskingObject

                
                self.DataSetModel.append(ds)
                self.DataSetModel.layoutChanged.emit()
                self.update()
                self.addProgressBarValue(1)
        else:
            self.setProgressBarLabelText('')
        DataFileListInfos = loadSetting(settingsFile,'infos')
        if not DataFileListInfos is None:
            self.DataFileInfoModel.infos = DataFileListInfos

        self.loadGuiSettings(file=settingsFile)
        self.loadLineEdits(file=settingsFile)
        self.loadRadioButtons(file=settingsFile)
        self.loadSpinBoxes(file=settingsFile)
        self.loadCheckBoxes(file=settingsFile)
        self.loadPredictionSettings(file=settingsFile)
        self.DataSetModel.layoutChanged.emit()
        self.DataFileInfoModel.layoutChanged.emit()
        self.DataFileModel.updateCurrentDataSetIndex()
        self.update()


        self.loadedGuiSettings = self.generateCurrentGuiSettings()
        return True

    def guiSettings(self):
        boxStates = [b.state for b in self.boxContainers]
        settingsDict = {'boxStates':boxStates,'colormap':self.colormap}
        if not hasattr(self,'BraggListWindow'):
           settingsDict['braggPoints'] = self.braggPoints
        else:
            settingsDict['braggPoints'] = self.BraggListWindow.BraggListModel.data
        return settingsDict
        
    def loadGuiSettings(self,file=None):
        
        if file is None:
            file = self.settingsFile
        
        guiSettings = loadSetting(file,'guiSettings')
        if not guiSettings is None:
            boxStates = guiSettings['boxStates']
            
            if not boxStates is None:
                for box,value in zip(self.boxContainers,boxStates):
                    try:
                        if box.state != value:
                            box.on_pressed()
                    except AttributeError:
                        pass

            if 'braggPoints' in guiSettings:
                self.braggPoints = guiSettings['braggPoints']
                if hasattr(self,'BraggListWindow'):
                    self.BraggListWindow.BraggListModel.data = self.braggPoints.copy()
                    self.BraggListWindow.BraggListModel.layoutChanged.emit()
            if 'colormap' in guiSettings:
                self.colormap = guiSettings['colormap']
            else:
                self.colormap = 'viridis'
        

    def loadLineEdits(self,file=None):
        if file is None:
            file = self.settingsFile
        lineEditValueString = loadSetting(file,'lineEdits')
        if not lineEditValueString is None:
            if isinstance(lineEditValueString,str):
                print('Please save a new gui state to comply with the new version')
                return
            for item,value in lineEditValueString.items():
                try:
                    getattr(self.ui,item).setText(value)
                except AttributeError:
                    pass

    def loadRadioButtons(self,file=None):
        if file is None:
            file = self.settingsFile
        radioButtonString = loadSetting(file,'radioButtons')
        if not radioButtonString is None:
            if isinstance(radioButtonString,str):
                print('Please save a new gui state to comply with the new version')
                return
            for item,value in radioButtonString.items():
                try:
                    getattr(self.ui,item).setChecked(value)
                except AttributeError:
                    pass

    def loadPredictionSettings(self,file=None):
        if file is None:
            file = self.settingsFile
        predictionSettings = loadSetting(file,'predictionSettings')
        if not predictionSettings is None:
            if isinstance(predictionSettings,str):
                print('Please save a new gui state to comply with the new version')
                return
            self.predictionSettings = predictionSettings

    def loadSpinBoxes(self,file=None):
        if file is None:
            file = self.settingsFile
        spinBoxValueString = loadSetting(file,'spinBoxes')
        if not spinBoxValueString is None:
            if isinstance(spinBoxValueString,str):
                print('Please save a new gui state to comply with the new version')
                return
            for item,value in spinBoxValueString.items():
                try:
                    getattr(self.ui,item).setValue(value)
                except AttributeError:
                    pass
    
    def loadCheckBoxes(self,file=None):
        if file is None:
            file = self.settingsFile
        checkBoxString = loadSetting(file,'checkBoxes')
        if not checkBoxString is None:
            if isinstance(checkBoxString,str):
                print('Please save a new gui state to comply with the new version')
                return
            for item,value in checkBoxString.items():
                try:
                    getattr(self.ui,item).setChecked(value)
                except AttributeError:
                    pass

    def getCurrentDirectory(self):
        return self.currentFolder

    def setCurrentDirectory(self,folder):
        self.currentFolder = folder
        
    def writeToStatus(self,text):
        self.log.append(text)

    def _clearStatus(self):
        self.log.clear()

    @ProgressBarDecoratorArguments(runningText='Clear Log',completedText='Log Cleared')
    def clearStatus(self):
        self._clearStatus()
        return True
    

    def resetProgressBarTimed(self):
        if self.current_timer:
            self.current_timer.stop()
        self.current_timer = QtCore.QTimer()
        self.current_timer.timeout.connect(self.resetProgressBar)
        self.current_timer.setSingleShot(True)
        self.current_timer.start(3000)

    #def changeTheme(self,name):
    #    if not name in themes.keys():
    #        raise AttributeError('Theme name not recognized. Got {}, but allowed are: '.format(name),', '.join(themes.keys()))
    #    app = QtWidgets.QApplication.instance()
    #    self.theme = name
    #    themes[name](app)
    #    #palette = app.palette()
    #    #print('Palette:',palette)
    #    #for view in self.views:
    #    #    view.setPalette(palette)


    def settingsDialog(self):
        # Get infos from DataFileInfoModel
        
        dataFileInfoModelPossibleSettings,dataFileInfoModelInitial = self.DataFileInfoModel.settingsDialog()
        # Create a widget holding check boxes for all possible settings
        dFIMLayout = QtWidgets.QVBoxLayout() # Outer most layout
        dFIMLayoutScrollArea = QtWidgets.QScrollArea() # scrollarea layout
        dummyWidget = QtWidgets.QWidget() # Dummy widget to insert layout into scroll area

        dFIMLayout_settings = QtWidgets.QVBoxLayout()
        dFIMTitleLabel = QtWidgets.QLabel(text='DataFile intormation \nSelect infos to be shown for selected file(s)')
        dFIMTitleLabel.setAlignment(QtCore.Qt.AlignCenter)
        # Add title to layout
        dFIMLayout.addWidget(dFIMTitleLabel)

        # make check boxes for all settings
        dFIMcheckBoxes = []
        for setting in dataFileInfoModelPossibleSettings.values():
            checkBox = QtWidgets.QCheckBox()
            dFIMcheckBoxes.append(checkBox)
            name = setting.location
            checkBox.setText(name)
            checkBox.setChecked(setting in dataFileInfoModelInitial)
            dFIMLayout_settings.addWidget(checkBox)
        dummyWidget.setLayout(dFIMLayout_settings)
        dFIMLayoutScrollArea.setWidget(dummyWidget)
        
        dFIMLayout.addWidget(dFIMLayoutScrollArea)
        # Overview: dFIMLayout -> dFIMLayoutScrollArea -> dummyWidget -> dFIMLayout_settings -> checkBox
        #     Type:   layout             widget              widget            layout            widget


        # accept function arguments: self (dialog), layout which was passed in
        def dFIMAcceptFunction(self,layout,possibleSettings=dataFileInfoModelPossibleSettings):
            self.dMFIASettings = []
            for idx,setting in enumerate(possibleSettings.values()): # Loop through all the possible settings
                box = layout.itemAt(1).widget().widget().layout().itemAt(idx).widget() # Skip 0 as it is a QLabel
                if box.isChecked():# If checked add the corresponding setting to list of loaded settings
                    self.dMFIASettings.append(setting.location)


        # Create layout for gui settings
        guiSettingsLayout = QtWidgets.QVBoxLayout()
        labelTitle = QtWidgets.QLabel(text = 'Color map:')

        cmapBox = QtWidgets.QComboBox()
        cmaps = plt.colormaps()
        current = cmaps.index(self.colormap)
        cmapBox.addItems(cmaps)
        cmapBox.setCurrentIndex(current)
        
        guiSettingsLayout.addWidget(labelTitle)
        guiSettingsLayout.addWidget(cmapBox)


        # Create radiobuttons
        
        def guiSettingsAcceptFunction(self,layout,guiWindow=self):
            cmapBox = layout.itemAt(1).widget()
            guiWindow.colormap = cmapBox.currentText()
            
 
        # settings holds a list of possible settings for all setting fields
        layouts = [guiSettingsLayout,dFIMLayout]
        acceptFunctions = [guiSettingsAcceptFunction,dFIMAcceptFunction]
        dialog = settingsBoxDialog(layouts=layouts,acceptFunctions=acceptFunctions)
        dialog.setWindowIcon(QtGui.QIcon(self.AppContext.get_resource('Icons/Own/settings.png')))
        dialog.resize(dialog.sizeHint())
        
        
        
        if dialog.exec_(): # Execute the dialog
            self.DataFileInfoModel.infos = dialog.dMFIASettings # update settings
            self.DataFileInfoModel.layoutChanged.emit()
        else:
            return
            
    def molarMassTool(self):
        molecularCalculationManager = MolecularCalculationManager(parent=None,guiWindow=self)
        self.windows.append(molecularCalculationManager)
        molecularCalculationManager.show()
        

    def neutronCalculationTool(self):
        calculatorManager = CalculatorManager(parent=None,guiWindow=self)
        self.windows.append(calculatorManager)
        calculatorManager.show()

    def absolutNormalizationTool(self):
        absolutNormalizationWindow = NormalizationManager(parent=None,guiWindow=self)
        self.windows.append(absolutNormalizationWindow)
        absolutNormalizationWindow.show()

    def predictionTool(self):
        predictionToolWindow = PredictionToolManager(parent=None,guiWindow=self)
        self.windows.append(predictionToolWindow)
        predictionToolWindow.show()

    def electronicLogbookTool(self):
        electronicLogBook = ElectronicLogBookManager(parent=None,guiWindow=self)
        self.windows.append(electronicLogBook)
        electronicLogBook.show()

    def subtractionManager(self):
        subtractionManager = SubtractionManager(parent=None,guiWindow=self)
        self.windows.append(subtractionManager)
        subtractionManager.show()

    def getBraggPoints(self):
        if hasattr(self,'BraggListWindow'):
            return self.BraggListWindow.getData()
        else:
            return self.braggPoints
    

    def openBraggListWindow(self):
        if hasattr(self,'BraggListWindow'): # If a window is open, use it
            if self.BraggListWindow.isVisible(): # if not visible, the window was closed
                getattr(self.BraggListWindow,'raise')()
                self.BraggListWindow.activateWindow()
            else:
                if self.braggPoints is None:
                    bL = []
                else:
                    bL = self.braggPoints.copy()
                self.BraggListWindow = BraggListManager.BraggListManager(BraggList=bL,guiWindow = self)
                self.BraggListWindow.BraggListModel.layoutChanged.connect(self.View3D_changed_CurratAxeList_function)

                #self.braggPoints = self.BraggListWindow.BraggListModel.data
            self.windows.append(self.BraggListWindow)
            self.BraggListWindow.show()

        else:
            if self.braggPoints is None:
                bL = []
            else:
                bL = self.braggPoints.copy()
            self.BraggListWindow = BraggListManager.BraggListManager(BraggList=bL,guiWindow = self)
            self.BraggListWindow.BraggListModel.layoutChanged.connect(self.View3D_changed_CurratAxeList_function)
            self.windows.append(self.BraggListWindow)
            self.BraggListWindow.show()

class settingsBoxDialog(QtWidgets.QDialog):

    def __init__(self, layouts, acceptFunctions, *args, **kwargs):
        super(settingsBoxDialog, self).__init__(*args, **kwargs)
        
        self.setWindowTitle("Settings")
        self.acceptFunctions = acceptFunctions
        self.layouts = layouts

        self.layout = QtWidgets.QVBoxLayout()
        
        for layout in layouts:
            self.layout.addLayout(layout)
        
        
        QBtn = QtWidgets.QDialogButtonBox.Ok | QtWidgets.QDialogButtonBox.Cancel
        
        self.buttonBox = QtWidgets.QDialogButtonBox(QBtn)
        self.buttonBox.accepted.connect(self.accept)
        self.buttonBox.rejected.connect(self.reject)

        
        self.layout.addWidget(self.buttonBox)
        self.setLayout(self.layout)

    def accept(self): # the accept button has been pressed
        for aFunc,layout in zip(self.acceptFunctions,self.layouts):
            aFunc(self,layout)
        return super(settingsBoxDialog,self).accept()

    def reject(self):
        return super(settingsBoxDialog,self).reject()



def updateSplash(splash,originalTime,updateInterval,padding='\n'*7+20*' '):
    currentTime = datetime.datetime.now()
    points = int(1000.0*(currentTime-originalTime).total_seconds()/updateInterval)+1

    alignment = QtCore.Qt.AlignTop# | QtCore.Qt.AlignHCenter
    splash.showMessage(padding+'Loading MJOLNIRGui'+'.'*points,color=QtGui.QColor(255,255,255),alignment=alignment)
    #QTimer.singleShot(1000, updateSplash(splash,points+1) )
    QtWidgets.QApplication.processEvents()

def main():
    try:
        import AppContextEmulator
    except ImportError:
        from MJOLNIRGui.src.main.python import AppContextEmulator
        

    app = QtWidgets.QApplication(sys.argv) # Passing command line arguments to app

    appEmu = AppContextEmulator.AppContextEmulator(__file__)

    splash = QtWidgets.QSplashScreen(QtGui.QPixmap(appEmu.get_resource('splash.png')))                                    
    splash.show()

    timer = QtCore.QTimer() 

    # adding action to timer 
    updateInterval = 400 # ms
    originalTime = datetime.datetime.now()
    updater = lambda:updateSplash(splash,originalTime=originalTime,updateInterval=updateInterval)
    updater()
    timer.timeout.connect(updater) 

    # update the timer every updateInterval 
    timer.start(updateInterval)
    

    window = MJOLNIRMainWindow(appEmu) # This window has to be closed for app to end
    splash.finish(window)
    window.show()
    timer.stop()

    if len(sys.argv)==2:
        window.loadGui(presetFileLocation=sys.argv[1])

    app.exec_() 

if __name__ == '__main__':
    main()

    


# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'main.ui'
#
# Created by: PyQt5 UI code generator 5.15.6
#
# WARNING: Any manual changes made to this file will be lost when pyuic5 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_MainWindow(object):
    def setupUi(self, MainWindow):
        MainWindow.setObjectName("MainWindow")
        MainWindow.resize(800, 950)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(MainWindow.sizePolicy().hasHeightForWidth())
        MainWindow.setSizePolicy(sizePolicy)
        MainWindow.setMinimumSize(QtCore.QSize(800, 340))
        MainWindow.setMaximumSize(QtCore.QSize(16777215, 16777215))
        MainWindow.setWindowOpacity(1.0)
        self.centralwidget = QtWidgets.QWidget(MainWindow)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Expanding)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.centralwidget.sizePolicy().hasHeightForWidth())
        self.centralwidget.setSizePolicy(sizePolicy)
        self.centralwidget.setObjectName("centralwidget")
        self.verticalLayout_2 = QtWidgets.QVBoxLayout(self.centralwidget)
        self.verticalLayout_2.setContentsMargins(0, 0, 0, 0)
        self.verticalLayout_2.setSpacing(0)
        self.verticalLayout_2.setObjectName("verticalLayout_2")
        self.scrollArea = QtWidgets.QScrollArea(self.centralwidget)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Expanding)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.scrollArea.sizePolicy().hasHeightForWidth())
        self.scrollArea.setSizePolicy(sizePolicy)
        self.scrollArea.setLineWidth(0)
        self.scrollArea.setVerticalScrollBarPolicy(QtCore.Qt.ScrollBarAlwaysOn)
        self.scrollArea.setHorizontalScrollBarPolicy(QtCore.Qt.ScrollBarAlwaysOff)
        self.scrollArea.setWidgetResizable(True)
        self.scrollArea.setObjectName("scrollArea")
        self.scrollAreaWidgetContents = QtWidgets.QWidget()
        self.scrollAreaWidgetContents.setGeometry(QtCore.QRect(0, 0, 781, 822))
        self.scrollAreaWidgetContents.setObjectName("scrollAreaWidgetContents")
        self.verticalLayout_4 = QtWidgets.QVBoxLayout(self.scrollAreaWidgetContents)
        self.verticalLayout_4.setContentsMargins(0, 0, 0, 0)
        self.verticalLayout_4.setObjectName("verticalLayout_4")
        self.verticalLayout = QtWidgets.QVBoxLayout()
        self.verticalLayout.setSizeConstraint(QtWidgets.QLayout.SetNoConstraint)
        self.verticalLayout.setSpacing(6)
        self.verticalLayout.setObjectName("verticalLayout")
        self.fixedOpen = QtWidgets.QWidget(self.scrollAreaWidgetContents)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.fixedOpen.sizePolicy().hasHeightForWidth())
        self.fixedOpen.setSizePolicy(sizePolicy)
        self.fixedOpen.setMinimumSize(QtCore.QSize(0, 300))
        self.fixedOpen.setObjectName("fixedOpen")
        self.verticalLayout.addWidget(self.fixedOpen)
        self.collapsibleContainer = QtWidgets.QWidget(self.scrollAreaWidgetContents)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Expanding)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.collapsibleContainer.sizePolicy().hasHeightForWidth())
        self.collapsibleContainer.setSizePolicy(sizePolicy)
        self.collapsibleContainer.setStyleSheet("")
        self.collapsibleContainer.setObjectName("collapsibleContainer")
        self.verticalLayout.addWidget(self.collapsibleContainer)
        self.verticalLayout_4.addLayout(self.verticalLayout)
        self.scrollArea.setWidget(self.scrollAreaWidgetContents)
        self.verticalLayout_2.addWidget(self.scrollArea)
        self.verticalLayout_3 = QtWidgets.QVBoxLayout()
        self.verticalLayout_3.setContentsMargins(10, 6, 10, -1)
        self.verticalLayout_3.setSpacing(10)
        self.verticalLayout_3.setObjectName("verticalLayout_3")
        self.progressBar = QtWidgets.QProgressBar(self.centralwidget)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.progressBar.sizePolicy().hasHeightForWidth())
        self.progressBar.setSizePolicy(sizePolicy)
        self.progressBar.setProperty("value", 0)
        self.progressBar.setObjectName("progressBar")
        self.verticalLayout_3.addWidget(self.progressBar)
        self.horizontalLayout = QtWidgets.QHBoxLayout()
        self.horizontalLayout.setContentsMargins(0, -1, 0, -1)
        self.horizontalLayout.setSpacing(10)
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.textBrowser = QtWidgets.QTextBrowser(self.centralwidget)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.textBrowser.sizePolicy().hasHeightForWidth())
        self.textBrowser.setSizePolicy(sizePolicy)
        self.textBrowser.setMaximumSize(QtCore.QSize(16777215, 45))
        self.textBrowser.setObjectName("textBrowser")
        self.horizontalLayout.addWidget(self.textBrowser)
        self.log_reset_btn = QtWidgets.QPushButton(self.centralwidget)
        self.log_reset_btn.setObjectName("log_reset_btn")
        self.horizontalLayout.addWidget(self.log_reset_btn)
        self.verticalLayout_3.addLayout(self.horizontalLayout)
        self.verticalLayout_2.addLayout(self.verticalLayout_3)
        MainWindow.setCentralWidget(self.centralwidget)
        self.menubar = QtWidgets.QMenuBar(MainWindow)
        self.menubar.setGeometry(QtCore.QRect(0, 0, 800, 21))
        self.menubar.setObjectName("menubar")
        self.menuFile = QtWidgets.QMenu(self.menubar)
        self.menuFile.setObjectName("menuFile")
        self.menuHelp = QtWidgets.QMenu(self.menubar)
        self.menuHelp.setObjectName("menuHelp")
        self.menuTools = QtWidgets.QMenu(self.menubar)
        self.menuTools.setObjectName("menuTools")
        MainWindow.setMenuBar(self.menubar)
        self.statusbar = QtWidgets.QStatusBar(MainWindow)
        self.statusbar.setObjectName("statusbar")
        MainWindow.setStatusBar(self.statusbar)
        self.actionSave_GUI_state = QtWidgets.QAction(MainWindow)
        self.actionSave_GUI_state.setObjectName("actionSave_GUI_state")
        self.actionLoad_GUI_state = QtWidgets.QAction(MainWindow)
        self.actionLoad_GUI_state.setObjectName("actionLoad_GUI_state")
        self.actionGenerate_View3d_script = QtWidgets.QAction(MainWindow)
        self.actionGenerate_View3d_script.setObjectName("actionGenerate_View3d_script")
        self.actionGenerate_QELine_script = QtWidgets.QAction(MainWindow)
        self.actionGenerate_QELine_script.setObjectName("actionGenerate_QELine_script")
        self.actionGenerate_QPlane_script = QtWidgets.QAction(MainWindow)
        self.actionGenerate_QPlane_script.setObjectName("actionGenerate_QPlane_script")
        self.actionGenerate_1d_script = QtWidgets.QAction(MainWindow)
        self.actionGenerate_1d_script.setObjectName("actionGenerate_1d_script")
        self.actionExit = QtWidgets.QAction(MainWindow)
        self.actionExit.setObjectName("actionExit")
        self.actionOpen_mask_gui = QtWidgets.QAction(MainWindow)
        self.actionOpen_mask_gui.setObjectName("actionOpen_mask_gui")
        self.actionLoad_mask = QtWidgets.QAction(MainWindow)
        self.actionLoad_mask.setObjectName("actionLoad_mask")
        self.actionAbout = QtWidgets.QAction(MainWindow)
        self.actionAbout.setObjectName("actionAbout")
        self.actionHelp = QtWidgets.QAction(MainWindow)
        self.actionHelp.setObjectName("actionHelp")
        self.actionChange_Theme = QtWidgets.QAction(MainWindow)
        self.actionChange_Theme.setObjectName("actionChange_Theme")
        self.actionSettings = QtWidgets.QAction(MainWindow)
        self.actionSettings.setObjectName("actionSettings")
        self.actionClose_Windows = QtWidgets.QAction(MainWindow)
        self.actionClose_Windows.setObjectName("actionClose_Windows")
        self.actionPrediction_Tool = QtWidgets.QAction(MainWindow)
        self.actionPrediction_Tool.setObjectName("actionPrediction_Tool")
        self.actionCalculate_Molecular_Weight = QtWidgets.QAction(MainWindow)
        self.actionCalculate_Molecular_Weight.setObjectName("actionCalculate_Molecular_Weight")
        self.actionNeutron_Calculations = QtWidgets.QAction(MainWindow)
        self.actionNeutron_Calculations.setObjectName("actionNeutron_Calculations")
        self.actionGenerate_Normalization = QtWidgets.QAction(MainWindow)
        self.actionGenerate_Normalization.setObjectName("actionGenerate_Normalization")
        self.actionElectronic_Logbook = QtWidgets.QAction(MainWindow)
        self.actionElectronic_Logbook.setObjectName("actionElectronic_Logbook")
        self.actionSubtraction_Of_DataSets = QtWidgets.QAction(MainWindow)
        self.actionSubtraction_Of_DataSets.setObjectName("actionSubtraction_Of_DataSets")
        self.actionSubtraction = QtWidgets.QAction(MainWindow)
        self.actionSubtraction.setObjectName("actionSubtraction")
        self.actionCheck_Update = QtWidgets.QAction(MainWindow)
        self.actionCheck_Update.setObjectName("actionCheck_Update")
        self.action_masking_gui = QtWidgets.QAction(MainWindow)
        self.action_masking_gui.setObjectName("action_masking_gui")
        self.menuFile.addAction(self.actionSave_GUI_state)
        self.menuFile.addAction(self.actionLoad_GUI_state)
        self.menuFile.addSeparator()
        self.menuFile.addAction(self.actionGenerate_View3d_script)
        self.menuFile.addAction(self.actionGenerate_QELine_script)
        self.menuFile.addAction(self.actionGenerate_QPlane_script)
        self.menuFile.addAction(self.actionGenerate_1d_script)
        self.menuFile.addSeparator()
        self.menuFile.addAction(self.actionClose_Windows)
        self.menuFile.addAction(self.actionSettings)
        self.menuFile.addAction(self.actionExit)
        self.menuHelp.addAction(self.actionAbout)
        self.menuHelp.addAction(self.actionHelp)
        self.menuHelp.addAction(self.actionSubtraction)
        self.menuTools.addAction(self.action_masking_gui)
        self.menuTools.addAction(self.actionPrediction_Tool)
        self.menuTools.addAction(self.actionCalculate_Molecular_Weight)
        self.menuTools.addAction(self.actionNeutron_Calculations)
        self.menuTools.addAction(self.actionGenerate_Normalization)
        self.menuTools.addAction(self.actionElectronic_Logbook)
        self.menuTools.addAction(self.actionSubtraction_Of_DataSets)
        self.menubar.addAction(self.menuFile.menuAction())
        self.menubar.addAction(self.menuTools.menuAction())
        self.menubar.addAction(self.menuHelp.menuAction())

        self.retranslateUi(MainWindow)
        QtCore.QMetaObject.connectSlotsByName(MainWindow)

    def retranslateUi(self, MainWindow):
        _translate = QtCore.QCoreApplication.translate
        MainWindow.setWindowTitle(_translate("MainWindow", "MJOLNIRGui"))
        self.fixedOpen.setToolTip(_translate("MainWindow", "Manage datasets and their datafiles"))
        self.progressBar.setToolTip(_translate("MainWindow", "Current progress of action"))
        self.progressBar.setStatusTip(_translate("MainWindow", "Current progress of action"))
        self.textBrowser.setToolTip(_translate("MainWindow", "Log of messages from the gui"))
        self.textBrowser.setStatusTip(_translate("MainWindow", "Log of messages from the gui"))
        self.log_reset_btn.setToolTip(_translate("MainWindow", "Clear log"))
        self.log_reset_btn.setStatusTip(_translate("MainWindow", "Clear log"))
        self.log_reset_btn.setText(_translate("MainWindow", "Clear"))
        self.menuFile.setTitle(_translate("MainWindow", "File"))
        self.menuHelp.setTitle(_translate("MainWindow", "Help"))
        self.menuTools.setTitle(_translate("MainWindow", "Tools"))
        self.actionSave_GUI_state.setText(_translate("MainWindow", "Save GUI state"))
        self.actionLoad_GUI_state.setText(_translate("MainWindow", "Load GUI state"))
        self.actionGenerate_View3d_script.setText(_translate("MainWindow", "Generate View3D script"))
        self.actionGenerate_QELine_script.setText(_translate("MainWindow", "Generate QE Cut script"))
        self.actionGenerate_QPlane_script.setText(_translate("MainWindow", "Generate QPlane script"))
        self.actionGenerate_1d_script.setText(_translate("MainWindow", "Generate 1D script"))
        self.actionExit.setText(_translate("MainWindow", "Exit"))
        self.actionOpen_mask_gui.setText(_translate("MainWindow", "Open Mask GUI"))
        self.actionLoad_mask.setText(_translate("MainWindow", "Load Mask"))
        self.actionAbout.setText(_translate("MainWindow", "About"))
        self.actionHelp.setText(_translate("MainWindow", "Help"))
        self.actionChange_Theme.setText(_translate("MainWindow", "Change Theme"))
        self.actionSettings.setText(_translate("MainWindow", "Settings"))
        self.actionClose_Windows.setText(_translate("MainWindow", "Close Windows"))
        self.actionPrediction_Tool.setText(_translate("MainWindow", "Prediction Tool"))
        self.actionCalculate_Molecular_Weight.setText(_translate("MainWindow", "Calculate Molecular Weight"))
        self.actionNeutron_Calculations.setText(_translate("MainWindow", "Neutron Calculations"))
        self.actionGenerate_Normalization.setText(_translate("MainWindow", "Generate Normalization"))
        self.actionElectronic_Logbook.setText(_translate("MainWindow", "Electronic Logbook"))
        self.actionSubtraction_Of_DataSets.setText(_translate("MainWindow", "Subtraction of DataSets"))
        self.actionSubtraction.setText(_translate("MainWindow", "Subtraction"))
        self.actionCheck_Update.setText(_translate("MainWindow", "Check Update"))
        self.action_masking_gui.setText(_translate("MainWindow", "Masking Gui"))



import sys

from matplotlib.pyplot import polar
sys.path.append('..')

try:
    from MJOLNIRGui.src.main.python._tools import ProgressBarDecoratorArguments,loadUI
    import MJOLNIRGui.src.main.python._tools as _GUItools
    from MJOLNIRGui.src.main.python.DataModels import MatplotlibFigureList,MatplotlibFigureListDelegate
except ImportError:
    from _tools import ProgressBarDecoratorArguments,loadUI
    import _tools as _GUItools
    from DataModels import MatplotlibFigureList,MatplotlibFigureListDelegate
from os import path
from PyQt5 import QtWidgets, uic
import numpy as np
from MJOLNIR.Data.DraggableShapes import extractCut1DPropertiesRectanglePerpendicular,extractCut1DPropertiesRectangleHorizontal, extractCut1DPropertiesRectangleVertical


@ProgressBarDecoratorArguments(runningText='Generating QELine plot',completedText='QELine plot generated')
def QELine_plot_button_function(self):    
    # First check if we have data, otherwise convert data
    if not self.stateMachine.requireStateByName('Converted'):
        return False
    
    ds = self.DataSetModel.getCurrentDataSet()
    if len(ds.convertedFiles)==0:
        self.DataSet_convertData_button_function()
        
    rlu = self.ui.QELine_SelectUnits_RLU_radioButton.isChecked()
    # Get the Q points
    HStart=float(self.ui.QELine_HStart_lineEdit.text())
    KStart=float(self.ui.QELine_KStart_lineEdit.text())
    
    HEnd=float(self.ui.QELine_HEnd_lineEdit.text())
    KEnd=float(self.ui.QELine_KEnd_lineEdit.text())
    if rlu:
        LStart=float(self.ui.QELine_LStart_lineEdit.text())
        LEnd=float(self.ui.QELine_LEnd_lineEdit.text())
    
            
    if rlu:
        Q1 = np.array([HStart,KStart,LStart])
        Q2 = np.array([HEnd,KEnd,LEnd])
    else:
        Q1 = np.array([HStart,KStart])
        Q2 = np.array([HEnd,KEnd])
   
    # Collect them into one array
    QPoints = np.array([Q1,Q2])
    
    # Define orthogonal width and minimum pixel size along Q-cut
    width = float(self.ui.QELine_Width_lineEdit.text())   
    minPixel = float(self.ui.QELine_MinPixel_lineEdit.text()) 

    # Define energy bins
    EMin=float(self.ui.QELine_EMin_lineEdit.text())
    EMax=float(self.ui.QELine_EMax_lineEdit.text())
    NPoints=int(self.ui.QELine_NPoints_lineEdit.text())       
    dE = (EMax-EMin)/(NPoints+1)
    #EnergyBins=np.linspace(EMin,EMax,NPoints)

    # Check various plot settings

    if self.ui.QELine_LogScale_checkBox.isChecked():
        log=True
    else:
        log=False  
        
    if self.ui.QELine_ConstantBins_checkBox.isChecked():
        constantBins=True
    else:
        constantBins=False              

    cut1DrectanglePerpendicularLocal = lambda viewer,dr:cut1DrectanglePerpendicular(self,viewer,dr)
    cut1DFunctionRectangleHorizontalLocal = lambda viewer,dr:cut1DFunctionRectangleHorizontal(self,viewer,dr)
    cut1DFunctionRectangleVerticalLocal = lambda viewer,dr: cut1DFunctionRectangleVertical(self,viewer,dr)
    #try:
    # Make plot
    ax,Data,bins = \
    ds.plotCutQE(q1=QPoints[0],q2=QPoints[1], width=width, \
                                    minPixel=minPixel, EMin=EMin,EMax=EMax,dE = dE,\
                                    rlu=rlu,cmap=self.colormap,cut1DFunctionRectanglePerpendicular=cut1DrectanglePerpendicularLocal,
                                    cut1DFunctionRectangleHorizontal=cut1DFunctionRectangleHorizontalLocal,
                                    cut1DFunctionRectangleVertical = cut1DFunctionRectangleVerticalLocal,
                                    outputFunction=self.writeToStatus) # log!constantBins=constantBins,

    currentFigure = ax.get_figure()
    self.figureListQELine.append(currentFigure)

    grid = self.ui.QELine_Grid_checkBox.isChecked()
    
    TitleText=self.ui.QELine_SetTitle_lineEdit.text()
    if TitleText == '':
        TitleText = self.ui.QELine_SetTitle_lineEdit.placeholderText()
    CAxisMin=float(self.ui.QELine_CAxisMin_lineEdit.text())
    CAxisMax=float(self.ui.QELine_CAxisMax_lineEdit.text())

    currentFigure.settings = {'QELine_SelectUnits_RLU_radioButton':rlu,
                    'QELine_SelectUnits_AA_radioButton': not rlu,
                    'QELine_Grid_checkBox':grid!=False,
                    'QELine_LogScale_checkBox':log,
                    'QELine_ConstantBins_checkBox':constantBins,
                    #'View3D_CurratAxe_checkBox':plotCurratAxe,
                    'QELine_HStart_lineEdit':HStart,
                    'QELine_KStart_lineEdit':KStart,
                    'QELine_LStart_lineEdit':LStart,
                    'QELine_HEnd_lineEdit':HEnd,
                    'QELine_KEnd_lineEdit':KEnd,
                    'QELine_LEnd_lineEdit':LEnd,
                    'QELine_Width_lineEdit':width,
                    'QELine_MinPixel_lineEdit':minPixel,
                    'QELine_EMax_lineEdit':EMax,
                    'QELine_EMin_lineEdit':EMin,
                    'QELine_NPoints_lineEdit':NPoints,
                    'QELine_SetTitle_lineEdit':TitleText,
                    'QELine_CAxisMax_lineEdit':CAxisMax,
                    'QELine_CAxisMin_lineEdit':CAxisMin}

    def setClosed(fig):
        fig.closed=True

    closeFunction = lambda event: setClosed(currentFigure)
    currentFigure.canvas.mpl_connect('close_event', closeFunction)
    # Make some final changes to the plot
    self.QELine=ax    
    
    currentFigure.set_size_inches(8,6)

    if self.ui.QELine_Grid_checkBox.isChecked():
        ax.grid(True)
    else:
        ax.grid(False)
    self.QELine_setCAxis_button_function()
    self.QELine_SetTitle_button_function()

    return True
    #except:
    #    _GUItools.dialog(text='QELine cut could not be made. Check the limits for the cut and try again!')
    #    return False

def QELine_Grid_checkBox_toggled_function(self):
    currentFigure = self.figureListQELine.getCurrentFigure()
    if not currentFigure is None:
        grid = self.ui.QELine_Grid_checkBox.isChecked()
        currentFigure.gca().grid(grid)
        currentFigure.settings['QELine_Grid_checkBox'] = grid!=False
        

        
def QELine_toggle_units_function(self):
    if self.ui.QELine_SelectUnits_RLU_radioButton.isChecked(): # changed to RLU
        # Change titles
        self.ui.QELine_HLabel.setText('H')
        self.ui.QELine_KLabel.setText('K')
        self.ui.QELine_LLabel.setText('L')
        self.ui.QELine_LStart_lineEdit.setEnabled(True)
        self.ui.QELine_LEnd_lineEdit.setEnabled(True)
    else: # Changing to AA
        self.ui.QELine_HLabel.setText('Qx')
        self.ui.QELine_KLabel.setText('Qy')
        self.ui.QELine_LLabel.setText('N/A')
        self.ui.QELine_LStart_lineEdit.setEnabled(False)
        self.ui.QELine_LEnd_lineEdit.setEnabled(False)

def QELine_setCAxis_button_function(self):       
    CAxisMin=float(self.ui.QELine_CAxisMin_lineEdit.text())
    CAxisMax=float(self.ui.QELine_CAxisMax_lineEdit.text())
    currentFigure = self.figureListQELine.getCurrentFigure()
    if not currentFigure is None:
        currentFigure.gca().set_clim(CAxisMin,CAxisMax)
        currentFigure.settings['QELine_CAxisMax_lineEdit'] = CAxisMax
        currentFigure.settings['QELine_CAxisMin_lineEdit'] =CAxisMin
        currentFigure.canvas.draw()

def QELine_SetTitle_button_function(self):
    currentFigure = self.figureListQELine.getCurrentFigure()
    if not currentFigure is None:
        TitleText=self.ui.QELine_SetTitle_lineEdit.text()
        if TitleText == '':
            TitleText = self.ui.QELine_SetTitle_lineEdit.placeholderText()
        currentFigure.gca().set_title(TitleText)
        currentFigure._title=TitleText
        currentFigure.settings['QELine_SetTitle_lineEdit'] = TitleText
        currentFigure.canvas.draw()

def QELine_DataSet_selectionChanged_function(self):
    ds = self.DataSetModel.getCurrentDataSet()
    if not ds is None:
        title = ds.name
    else:
        title = ''
    self.ui.QELine_SetTitle_lineEdit.setPlaceholderText(title)

def indexChanged(self,index):
    figure = self.figureListQELine.figures[index]
    if hasattr(figure,'settings'):
        for setting,value in figure.settings.items():
            if 'radio' in setting or 'checkBox' in setting:
                getattr(getattr(self.ui,setting),'setChecked')(value)
            else:
                getattr(getattr(self.ui,setting),'setText')(str(value))
    if not figure is None:
        self.ui.QELine_figureToCSV_pushButton.setEnabled(True)
    else:
        
        self.ui.QELine_figureToCSV_pushButton.setEnabled(False)

@ProgressBarDecoratorArguments(runningText='Save QELine to csv',completedText='QELine saved')
def saveToCSV(self):
    figure = self.figureListQELine.getCurrentFigure()
    ax = figure.gca()
    title = ax.title.get_text()
    location,_ = QtWidgets.QFileDialog.getSaveFileName(self,'Save QELine',str(title)+'.csv','CSV (*.csv)')
    if location is None or location == '':
        return False
    if not location.split('.')[-1] == 'csv':
        location = location+'.csv'
    
    ax.to_csv(location)
    return True

def cut1DrectanglePerpendicular(self,ax,dr):
    # If there is no sample, i.e. 1/AA plot
    if hasattr(ax,'sample'):
        sample = ax.sample
    else:
        sample = None

    # Extract the parameters
    rounding = 4 # Round to 4 digits
    parameters = extractCut1DPropertiesRectanglePerpendicular(dr.rect,sample)

    offset = ax.convertPlotAxisReal(parameters['center'][0]).T[0]
    
    del parameters['center'] # remove the 'center' as it is not allowed in plotCut1D

    # transform the orthogonal vector if needed 
    if not sample is None:
        orthogonalVector = sample.calculateQxQyToHKL(*ax.orthovec)
    else:
        orthogonalVector = ax.orthovec
    
    # find actual offset along current cut
    offset -= np.dot(orthogonalVector,offset)*orthogonalVector/np.dot(orthogonalVector.T,orthogonalVector)
    
    parameters['q1'] = [np.round(x,rounding) for x in ax.minOrthoPosition+offset]
    parameters['q2'] = [np.round(x,rounding) for x in ax.maxOrthoPosition+offset]

    parameters['minPixel'] = np.round(ax.minPixel,rounding)

    
    # Order of parameters needed is: ds,q1,q2,width,minPixel,EMax,EMin,cutQ,rlu
    self.interactiveCut = [ax.ds,parameters['q1'],parameters['q2'],
                           np.round(parameters['width'],rounding),parameters['minPixel'],np.round(parameters['Emax'],rounding),np.round(parameters['Emin'],rounding),True,parameters['rlu']]
    
    # Perform the cut and plot it
    self.Cut1D_plot_button_function()
    # Reset the interactiveCut flag
    self.interactiveCut = None

def cut1DFunctionRectangleHorizontal(self,ax,dr):
    rounding = 4
    if hasattr(ax,'sample'):
        sample = ax.sample
    else:
        sample = None
    parameters = extractCut1DPropertiesRectangleHorizontal(dr.rect,sample)
    
    # Convert center point into actual position in Q
    parameters['q1'] = [np.round(x,rounding) for x in ax.convertPlotAxisReal(parameters['q1'][0]).T[0]]
    parameters['q2'] = [np.round(x,rounding) for x in ax.convertPlotAxisReal(parameters['q2'][0]).T[0]]
    
    parameters['minPixel'] = np.round(ax.minPixel,rounding)
    parameters['width'] = np.round(ax.width,rounding)
    self.interactiveCut = [ax.ds,parameters['q1'],parameters['q2'],
                           parameters['width'],parameters['minPixel'],np.round(parameters['Emax'],rounding),np.round(parameters['Emin'],rounding),True,parameters['rlu']]
    
    # Perform the cut and plot it
    self.Cut1D_plot_button_function()
    # Reset the interactiveCut flag
    self.interactiveCut = None

def cut1DFunctionRectangleVertical(self,ax,dr):
    rounding = 4
    if hasattr(ax,'sample'):
        sample = ax.sample
    else:
        sample = None
    parameters = extractCut1DPropertiesRectangleVertical(dr.rect,sample)
    
    # Convert center point into actual position in Q
    parameters['q'] = [np.round(x,rounding) for x in ax.convertPlotAxisReal(parameters['q']).T[0]]
    parameters['q2'] = None
    
    parameters['minPixel'] = np.round(ax.minPixel,rounding)
    parameters['width'] = np.round(ax.width,rounding)
    self.interactiveCut = [ax.ds,parameters['q'],None,
                           parameters['width'],parameters['minPixel'],np.round(parameters['E2'],rounding),np.round(parameters['E1'],rounding),False,parameters['rlu']]
    
    # Perform the cut and plot it
    self.Cut1D_plot_button_function()
    # Reset the interactiveCut flag
    self.interactiveCut = None

QELineManagerBase, QELineManagerForm = loadUI('QELine.ui')

class QELineManager(QELineManagerBase, QELineManagerForm):
    def __init__(self, parent=None, guiWindow=None):
        super(QELineManager, self).__init__(parent)
        self.setupUi(self)
        self.guiWindow = guiWindow
        self.initQELineManager()

    def initQELineManager(self):
        self.guiWindow.QELine_plot_button_function = lambda: QELine_plot_button_function(self.guiWindow)
        self.guiWindow.QELine_setCAxis_button_function = lambda: QELine_setCAxis_button_function(self.guiWindow)
        self.guiWindow.QELine_SetTitle_button_function = lambda: QELine_SetTitle_button_function(self.guiWindow)
        self.guiWindow.QELine_toggle_units_function = lambda: QELine_toggle_units_function(self.guiWindow)
        self.guiWindow.QELine_DataSet_selectionChanged_function = lambda: QELine_DataSet_selectionChanged_function(self.guiWindow)
        self.guiWindow.QELine_Grid_checkBox_toggled_function = lambda: QELine_Grid_checkBox_toggled_function(self.guiWindow)
        self.guiWindow.QELine_indexChanged = lambda index: indexChanged(self.guiWindow,index)
        self.guiWindow.QELine_saveToCSV = lambda: saveToCSV(self.guiWindow)

        for key,value in self.__dict__.items():
            if 'QELine' in key:
                self.guiWindow.ui.__dict__[key] = value

        ## Set up figure manager. Is to be initialized between manager initialization but before their setup
        self.guiWindow.figureListQELine = MatplotlibFigureList(combobox=self.QELine_figureList_comboBox)
        self.guiWindow.figureList.append(self.guiWindow.figureListQELine)
        self.QELine_figureList_comboBox.setModel(self.guiWindow.figureListQELine)

        self.mplListDelegate = MatplotlibFigureListDelegate()

    def setup(self):
        
        self.guiWindow.ui.QELine_plot_button.clicked.connect(self.guiWindow.QELine_plot_button_function)
        self.guiWindow.ui.QELine_setCAxis_button.clicked.connect(self.guiWindow.QELine_setCAxis_button_function)
        self.guiWindow.ui.QELine_SetTitle_button.clicked.connect(self.guiWindow.QELine_SetTitle_button_function)
        self.guiWindow.ui.QELine_SelectUnits_RLU_radioButton.toggled.connect(self.guiWindow.QELine_toggle_units_function)

        self.guiWindow.ui.QELine_CAxisMax_lineEdit.returnPressed.connect(self.CAxisChanged)
        self.guiWindow.ui.QELine_CAxisMin_lineEdit.returnPressed.connect(self.CAxisChanged)
        self.guiWindow.ui.QELine_Grid_checkBox.toggled.connect(self.guiWindow.QELine_Grid_checkBox_toggled_function)
        self.guiWindow.ui.QELine_SetTitle_lineEdit.returnPressed.connect(self.TitleChanged)

        self.guiWindow.DataSetSelectionModel.selectionChanged.connect(self.guiWindow.QELine_DataSet_selectionChanged_function)
        self.guiWindow.DataSetModel.dataChanged.connect(self.guiWindow.QELine_DataSet_selectionChanged_function)

        self.guiWindow.figureListQELine.view.currentIndexChanged.connect(self.guiWindow.QELine_indexChanged)
        self.guiWindow.ui.QELine_figureList_comboBox.setItemDelegate(self.mplListDelegate)

        self.guiWindow.ui.QELine_figureToCSV_pushButton.clicked.connect(self.guiWindow.QELine_saveToCSV)

    def CAxisChanged(self):
        if self.guiWindow.ui.QELine_setCAxis_button.isEnabled():
            self.guiWindow.QELine_setCAxis_button_function()

    def TitleChanged(self):
        if self.guiWindow.ui.QELine_SetTitle_button.isEnabled():
            self.guiWindow.QELine_SetTitle_button_function()



import sys
sys.path.append('..')

from MJOLNIR._tools import EnergyK,WavelengthK,WavelengthEnergy,EnergyWavelength,KEnergy,KWavelength
try:
    from _tools import loadUI
except ModuleNotFoundError:
    from MJOLNIRGui.src.main.python._tools import loadUI

from PyQt5 import QtGui

# Handles all functionality related to the CalculatorGeneralManager. 


CalculatorGeneralManagerBase, CalculatorGeneralManagerForm = loadUI('Calculator_General.ui')

       

def textChangedWavelength(manager):
    wavelength = manager.wavelength_spinBox.value()
    manager.q_spinBox.setValue(KWavelength(wavelength))
    manager.e_spinBox.setValue(EnergyWavelength(wavelength))

def textChangedEnergy(manager):
    E = manager.e_spinBox.value()
    manager.q_spinBox.setValue(KEnergy(E))
    manager.wavelength_spinBox.setValue(WavelengthEnergy(E))

def textChangedK(manager):
    k = manager.q_spinBox.value()
    manager.e_spinBox.setValue(EnergyK(k))
    manager.wavelength_spinBox.setValue(WavelengthK(k))

def onFocus(self,event,others):
    for o in others:
        try:
            o.valueChanged.disconnect()
        except TypeError: # If no connection to remove
            pass
    self.valueChanged.connect(self.onChangeFunction)
    self.old_focusInEvent(event)




class CalculatorGeneralManager(CalculatorGeneralManagerBase, CalculatorGeneralManagerForm):
    def __init__(self, parent=None, guiWindow=None):
        super(CalculatorGeneralManager, self).__init__(parent)
        self.setupUi(self)
        self.guiWindow = guiWindow
        self.setWindowIcon(QtGui.QIcon(self.guiWindow.AppContext.get_resource('Icons/Own/calculator.png')))

        self.initCalculatorGeneralManager()
        
    def initCalculatorGeneralManager(self):    
        self.setup()
        

    def setup(self):
        # Add updating functions to be called when text is changed
        self.wavelength_spinBox.onChangeFunction = lambda: textChangedWavelength(self)
        self.e_spinBox.onChangeFunction = lambda: textChangedEnergy(self)
        self.q_spinBox.onChangeFunction = lambda: textChangedK(self)
        
        # Move default focusInEvent
        self.wavelength_spinBox.old_focusInEvent = self.wavelength_spinBox.focusInEvent
        self.e_spinBox.old_focusInEvent = self.e_spinBox.focusInEvent
        self.q_spinBox.old_focusInEvent = self.q_spinBox.focusInEvent

        # Update to new 
        self.wavelength_spinBox.focusInEvent= lambda event: onFocus(self.wavelength_spinBox,event,[self.e_spinBox,self.q_spinBox])
        self.e_spinBox.focusInEvent = lambda event: onFocus(self.e_spinBox,event,[self.wavelength_spinBox,self.q_spinBox])
        self.q_spinBox.focusInEvent = lambda event: onFocus(self.q_spinBox,event,[self.wavelength_spinBox,self.e_spinBox])
        



import sys
sys.path.append('..')

try:
    from MJOLNIRGui.src.main.python._tools import ProgressBarDecoratorArguments, loadUI
    from MJOLNIRGui.src.main.python.DataModels import MatplotlibFigureList,MatplotlibFigureListDelegate
except ImportError:
    from _tools import ProgressBarDecoratorArguments, loadUI
    from DataModels import MatplotlibFigureList,MatplotlibFigureListDelegate

from os import path
from PyQt5 import QtWidgets, uic
import numpy as np

from MJOLNIR.Data.DraggableShapes import extractCut1DPropertiesRectangle, extractCut1DPropertiesCircle

@ProgressBarDecoratorArguments(runningText='Generating QPlane plot',completedText='QPlane plotted')
def QPlane_plot_button_function(self):
    # Make plot
    if not self.stateMachine.requireStateByName('Converted'):
        return False
    ds = self.DataSetModel.getCurrentDataSet()
    if len(ds.convertedFiles)==0:
        self.DataSet_convertData_button_function()        
    
    # Check various plot settings
    if self.ui.QPlane_SelectUnits_RLU_radioButton.isChecked():
        rlu=True
    else:
        rlu=False        

    if self.ui.QPlane_LogScale_checkBox.isChecked():
        log=True
    else:
        log=False          

    EMin=float(self.ui.QPlane_EMin_lineEdit.text())
    EMax=float(self.ui.QPlane_EMax_lineEdit.text())
    xBinTolerance = float(self.ui.QPlane_xBinTolerance_lineEdit.text())           
    yBinTolerance = float(self.ui.QPlane_yBinTolerance_lineEdit.text())           

    cut1DFunctionLocalRectangle = lambda viewer,dr:cut1DFunctionRectangle(self,viewer,dr)
    cut1DFunctionLocalCircle = lambda viewer,dr:cut1DFunctionCircle(self,viewer,dr)

    Data,Bins,ax = ds.plotQPlane(EMin=EMin, EMax=EMax,xBinTolerance=xBinTolerance,yBinTolerance=yBinTolerance,log=log,rlu=rlu,cmap=self.colormap,
    cut1DFunctionRectangle = cut1DFunctionLocalRectangle, cut1DFunctionCircle=cut1DFunctionLocalCircle,outputFunction=self.writeToStatus)
    currentFigure = ax.get_figure()
    self.QPlane=ax   
    
    self.figureListQPlane.append(currentFigure) 
    
    
    currentFigure.colorbar(ax.pmeshs[0])
    currentFigure.set_size_inches(8,6)
    grid = self.ui.QPlane_Grid_checkBox.isChecked()
    ax.grid(grid)
    TitleText=self.ui.QPlane_SetTitle_lineEdit.text()
    if TitleText == '':
        TitleText = self.ui.QPlane_SetTitle_lineEdit.placeholderText()
    CAxisMin=float(self.ui.QPlane_CAxisMin_lineEdit.text())
    CAxisMax=float(self.ui.QPlane_CAxisMax_lineEdit.text())

    currentFigure.settings = {'QPlane_SelectUnits_RLU_radioButton':rlu,
                    'QPlane_SelectUnits_AA_radioButton': not rlu,
                    'QPlane_Grid_checkBox':grid!=False,
                    'QPlane_LogScale_checkBox':log,
                    #'View3D_CurratAxe_checkBox':plotCurratAxe,
                    'QPlane_xBinTolerance_lineEdit':xBinTolerance,
                    'QPlane_yBinTolerance_lineEdit':yBinTolerance,
                    'QPlane_EMax_lineEdit':EMax,
                    'QPlane_EMin_lineEdit':EMin,
                    'QPlane_SetTitle_lineEdit':TitleText,
                    'QPlane_CAxisMax_lineEdit':CAxisMax,
                    'QPlane_CAxisMin_lineEdit':CAxisMin,
                    'QPlane_SetTitle_lineEdit':TitleText}

    def setClosed(fig):
        fig.closed=True

    closeFunction = lambda event: setClosed(currentFigure)
    currentFigure.canvas.mpl_connect('close_event', closeFunction)
                    
    self.QPlane_setCAxis_button_function()
    self.QPlane_SetTitle_button_function()
    self.windows.append(self.QPlane.get_figure())

    return True
    

def QPlane_setCAxis_button_function(self):
    CAxisMin=float(self.ui.QPlane_CAxisMin_lineEdit.text())
    CAxisMax=float(self.ui.QPlane_CAxisMax_lineEdit.text())
    currentFigure = self.figureListQPlane.getCurrentFigure()
    if not currentFigure is None:
        currentFigure.gca().set_clim(CAxisMin,CAxisMax)
    
        currentFigure.settings['QPlane_CAxisMax_lineEdit'] = CAxisMax
        currentFigure.settings['QPlane_CAxisMin_lineEdit'] = CAxisMin
    
def QPlane_SetTitle_button_function(self):
    TitleText=self.ui.QPlane_SetTitle_lineEdit.text()        
    if TitleText == '':
        TitleText = self.ui.QPlane_SetTitle_lineEdit.placeholderText()
    
    currentFigure = self.figureListQPlane.getCurrentFigure()
    if not currentFigure is None:
        currentFigure._title = TitleText
        currentFigure.gca().set_title(TitleText)
        currentFigure.settings['QPlane_SetTitle_lineEdit'] = TitleText
    

def QPlane_Grid_checkBox_toggled_function(self):
    currentFigure = self.figureListQPlane.getCurrentFigure()
    if not currentFigure is None:
        grid = self.ui.QPlane_Grid_checkBox.isChecked()
        currentFigure.gca().grid(grid)
        currentFigure.settings['QPlane_Grid_checkBox'] = grid

def QPlane_DataSet_selectionChanged_function(self):
    ds = self.DataSetModel.getCurrentDataSet()
    if not ds is None:
        title = ds.name
    else:
        title = ''
    self.ui.QPlane_SetTitle_lineEdit.setPlaceholderText(title)

def indexChanged(self,index):
    figure = self.figureListQPlane.figures[index]
    if hasattr(figure,'settings'):
        for setting,value in figure.settings.items():
            if 'radio' in setting or 'checkBox' in setting:
                getattr(getattr(self.ui,setting),'setChecked')(value)
            else:
                getattr(getattr(self.ui,setting),'setText')(str(value))
    if not figure is None:
        self.ui.QPlane_figureToCSV_pushButton.setEnabled(True)
    else:
        
        self.ui.QPlane_figureToCSV_pushButton.setEnabled(False)

@ProgressBarDecoratorArguments(runningText='Save QPlane to csv',completedText='QPlane saved')
def saveToCSV(self):
    figure = self.figureListQPlane.getCurrentFigure()
    ax = figure.gca()
    title = ax.title.get_text()
    location,_ = QtWidgets.QFileDialog.getSaveFileName(self,'Save QPlane',str(title)+'.csv','CSV (*.csv)')
    if location is None or location == '':
        return False
    if not location.split('.')[-1] == 'csv':
        location = location+'.csv'
    
    ax.to_csv(location)
    return True

def cut1DFunctionRectangle(self,ax,dr):
    # If there is no sample, i.e. 1/AA plot
    if hasattr(ax,'sample'):
        sample = ax.sample
    else:
        sample = None

    # Extract the parameters
    rounding = 4 # Round to 4 digits
    parameters = extractCut1DPropertiesRectangle(dr.rect,sample,rounding = rounding)
    
    
    parameters['Emin']=ax.EMin
    parameters['Emax']=ax.EMax

    # Order of parameters needed is: ds,q1,q2,width,minPixel,EMax,EMin,cutQ,rlu
    self.interactiveCut = [ax.ds,parameters['q1'],parameters['q2'],
                           parameters['width'],parameters['minPixel'],ax.EMax,ax.EMin,True,parameters['rlu']]
    
    # Perform the cut and plot it
    self.Cut1D_plot_button_function()
    # Reset the interactiveCut flag
    self.interactiveCut = None
    
def cut1DFunctionCircle(self,ax,dr):
    # If there is no sample, i.e. 1/AA plot
    if hasattr(ax,'sample'):
        sample = ax.sample
    else:
        sample = None

    # Extract the parameters
    rounding = 4 # Round to 4 digits
    parameters = extractCut1DPropertiesCircle(dr.circ,ax.sample)
    parameters['E1'] = ax.ds.energy.min()
    parameters['E2'] = ax.ds.energy.max()
    parameters['minPixel'] = np.min([ax.EMax-ax.EMin,0.15]) # Choose the smaller of the actual plane or 0.150 meV
    
    parameters['Emin']=ax.EMin
    parameters['Emax']=ax.EMax

    # Order of parameters needed is: ds,q1,q2,width,minPixel,EMax,EMin,cutQ,rlu
    self.interactiveCut = [ax.ds,parameters['q'],None,
                           parameters['width'],parameters['minPixel'],parameters['E2'],parameters['E1'],False,parameters['rlu']]
    
    # Perform the cut and plot it
    self.Cut1D_plot_button_function()
    # Reset the interactiveCut flag
    self.interactiveCut = None

QPlaneManagerBase, QPlaneManagerForm = loadUI('QPlane.ui')

class QPlaneManager(QPlaneManagerBase, QPlaneManagerForm):
    def __init__(self, parent=None, guiWindow=None):
        super(QPlaneManager, self).__init__(parent)
        self.setupUi(self)
        self.guiWindow = guiWindow
        self.initQPlaneManager()

    def initQPlaneManager(self):
        self.guiWindow.QPlane_plot_button_function = lambda: QPlane_plot_button_function(self.guiWindow)
        self.guiWindow.QPlane_setCAxis_button_function = lambda: QPlane_setCAxis_button_function(self.guiWindow)
        self.guiWindow.QPlane_SetTitle_button_function = lambda: QPlane_SetTitle_button_function(self.guiWindow)
        self.guiWindow.QPlane_Grid_checkBox_toggled_function = lambda: QPlane_Grid_checkBox_toggled_function(self.guiWindow)
        self.guiWindow.QPlane_DataSet_selectionChanged_function = lambda: QPlane_DataSet_selectionChanged_function(self.guiWindow)
        self.guiWindow.QPlane_indexChanged = lambda index: indexChanged(self.guiWindow,index)
        self.guiWindow.QPlane_saveToCSV = lambda: saveToCSV(self.guiWindow)

        for key,value in self.__dict__.items():
            if 'QPlane' in key:
                self.guiWindow.ui.__dict__[key] = value

        self.guiWindow.figureListQPlane = MatplotlibFigureList(combobox=self.QPlane_figureList_comboBox)
        self.guiWindow.figureList.append(self.guiWindow.figureListQPlane)
        self.QPlane_figureList_comboBox.setModel(self.guiWindow.figureListQPlane)

        self.mplListDelegate = MatplotlibFigureListDelegate()

    def setup(self):
        
        self.guiWindow.ui.QPlane_plot_button.clicked.connect(self.guiWindow.QPlane_plot_button_function)
        self.guiWindow.ui.QPlane_setCAxis_button.clicked.connect(self.guiWindow.QPlane_setCAxis_button_function)
        self.guiWindow.ui.QPlane_SetTitle_button.clicked.connect(self.guiWindow.QPlane_SetTitle_button_function)

        self.guiWindow.ui.QPlane_CAxisMax_lineEdit.returnPressed.connect(self.CAxisChanged)
        self.guiWindow.ui.QPlane_CAxisMin_lineEdit.returnPressed.connect(self.CAxisChanged)

        self.guiWindow.ui.QPlane_Grid_checkBox.toggled.connect(self.guiWindow.QPlane_Grid_checkBox_toggled_function)
        self.guiWindow.ui.QPlane_SetTitle_lineEdit.returnPressed.connect(self.TitleChanged)

        self.guiWindow.DataSetSelectionModel.selectionChanged.connect(self.guiWindow.QPlane_DataSet_selectionChanged_function)
        self.guiWindow.DataSetModel.dataChanged.connect(self.guiWindow.QPlane_DataSet_selectionChanged_function)

        self.guiWindow.figureListQPlane.view.currentIndexChanged.connect(self.guiWindow.QPlane_indexChanged)
        self.guiWindow.ui.QPlane_figureList_comboBox.setItemDelegate(self.mplListDelegate)

        self.guiWindow.ui.QPlane_figureToCSV_pushButton.clicked.connect(self.guiWindow.QPlane_saveToCSV)

    def CAxisChanged(self):
        if self.guiWindow.ui.QPlane_setCAxis_button.isEnabled():
            self.guiWindow.QPlane_setCAxis_button_function()

    def TitleChanged(self):
        if self.guiWindow.ui.QPlane_SetTitle_button.isEnabled():
            self.guiWindow.QPlane_SetTitle_button_function()



from PyQt5 import QtCore, QtGui, QtWidgets
from PyQt5.QtCore import Qt

import pickle

import numpy as np
import sys
import warnings

from MJOLNIR import _tools as M_tools
import matplotlib.pyplot as plt
try:
    from MJOLNIRGui.src.main.python.MJOLNIR_Data import GuiDataFile
except ImportError:
    from  MJOLNIR_Data import GuiDataFile

from os import path
from collections import namedtuple
try:
    import MJOLNIRGui.src.main.python._tools
except ImportError:
    import _tools

class DataSetModel(QtCore.QAbstractListModel):

    dragDropFinished = QtCore.pyqtSignal()

    def __init__(self, *args, dataSets=None, DataSet_DataSets_listView=None,guiWindow=None, **kwargs):
        super(DataSetModel, self).__init__(*args, **kwargs)
        self.dataSets = dataSets or []
        self.DataSet_DataSets_listView = DataSet_DataSets_listView
        self.lastDroppedItems = []
        self.pendingRemoveRowsAfterDrop = False
        self.guiWindow = guiWindow

    def rowForItem(self, item):
        '''
        rowForItem method returns the row corresponding to the passed in item
        or None if no such item exists in the model
        '''
        try:
            row = self.dataSets.index(item)
        except ValueError:
            return None
        return row
        
    def data(self, index, role = Qt.ItemDataRole):
        if role == Qt.DisplayRole or role == QtCore.Qt.EditRole:
            text = self.dataSets[index.row()].name
            return text
        if role == Qt.ItemDataRole:
            return self.dataSets[index.row()]
        if role == Qt.DecorationRole:
            if not self.dataSets[index.row()]._maskingObject is None:
                return QtGui.QImage(self.guiWindow.AppContext.get_resource('Icons/Own/mask-open.png'))
            else:
                return None

        #    if status:
        #        return tick

    def getData(self,*args,**kwargs):
        return self.data(*args,**kwargs)

    def rowCount(self, index):
        return len(self.dataSets)

    def insertRows(self, row, count, index):
        if index.isValid():
            return False
        if count <= 0:
            return False
        # inserting 'count' empty rows starting at 'row'
        self.beginInsertRows(QtCore.QModelIndex(), row, row + count - 1)
        for i in range(0, count):
            self.dataSets.insert(row + i, None)
        self.endInsertRows()
        return True

    def removeRows(self, row, count, index):
        if index.isValid():
            return False
        if count <= 0:
            return False
        # num_rows = self.rowCount(QtCore.QModelIndex())
        self.beginRemoveRows(QtCore.QModelIndex(), row, row + count - 1)
        for i in range(count, 0, -1):
            self.dataSets.pop(row - i + 1)
        self.endRemoveRows()

        if self.pendingRemoveRowsAfterDrop:
            '''
            If we got here, it means this call to removeRows is the automatic
            'cleanup' action after drag-n-drop performed by Qt
            '''
            self.pendingRemoveRowsAfterDrop = False
            self.dragDropFinished.emit()

        return True

    def generateValidName(self,ds):
        name = ds.name
        while name in [DS.name for DS in self.dataSets if not DS is ds]: # name already exists.. This screws up drag/drop
            try:
                idx = int(name.split(' ')[-1])
            except ValueError:
                name = name+' 1'
            else:
                name = name[:-(len(str(idx))+1)] + ' ' + str(idx+1)
        return name

    def append(self,ds):
        ds.name = self.generateValidName(ds)

        if self.rowCount(None)>0:
            numbers = [d.idx for d in self.dataSets]
        else:
            numbers = [-1]
        ds.idx = np.max(numbers)+1
        self.dataSets.append(ds)
        self.selectLastDataSet()
        self.layoutChanged.emit()

    def delete(self,index):
        try:
            del self.dataSets[index.row()]
            self.layoutChanged.emit()
        except:
            pass
        QtWidgets.QApplication.processEvents()
        index = self.getCurrentDatasetIndex()
        
        if index is None:
            self.selectLastDataSet()
        else:
            if index.row()==self.rowCount(None):
                self.selectLastDataSet()
            else:
                idx = self.index(index.row(),0)# Hack to force index changed emit
                self.DataSet_DataSets_listView.setCurrentIndex(self.index(-1,0))
                self.DataSet_DataSets_listView.setCurrentIndex(idx)


    def item(self,index):
        if not index is None:
            return self.dataSets[index.row()]

    def setData(self, index, value, role=QtCore.Qt.EditRole):
        ds = self.item(index)
        if role == QtCore.Qt.EditRole:
            ds.name = value
            self.dataChanged.emit(index, index)
            return True
            
        return False

    def index(self, row, column, parent=None):
        if row < 0 or row >= self.rowCount(None):
            return QtCore.QModelIndex()
        return self.createIndex(row, column)

    def parent(self, index):
        return QtCore.QModelIndex()

    def supportedDropActions(self):
        return QtCore.Qt.MoveAction 
        

    def flags(self,index):
        if not index.isValid():
            return QtCore.Qt.ItemIsEnabled | QtCore.Qt.ItemIsDropEnabled
        return QtCore.Qt.ItemIsEditable | QtCore.Qt.ItemIsEnabled | QtCore.Qt.ItemIsSelectable | \
            QtCore.Qt.ItemIsDragEnabled | QtCore.Qt.ItemIsDropEnabled

    def mimeTypes(self):
        return ['mjolnirgui/datasetdragdrop.list']

    def mimeData(self, indexes):
        mimedata = QtCore.QMimeData()
        encoded_data = QtCore.QByteArray()
        stream = QtCore.QDataStream(encoded_data, QtCore.QIODevice.WriteOnly)
        for index in indexes:
            if index.isValid():
                item = self.data(index, QtCore.Qt.ItemDataRole)
                
        stream << QtCore.QByteArray(str(item.idx).encode('utf-8'))#pickle.dumps(item))
        mimedata.setData('mjolnirgui/datasetdragdrop.list', encoded_data)
        return mimedata

    def dropMimeData(self, data, action, row, column, parent):
        if action == QtCore.Qt.IgnoreAction:
            return True
        if not data.hasFormat('mjolnirgui/datasetdragdrop.list'):
            return False
        if column > 0:
            return False

        num_rows = self.rowCount(QtCore.QModelIndex())
        if num_rows <= 0:
            return False

        if row < 0:
            if parent.isValid():
                row = parent.row()
            else:
                return False

        encoded_data = data.data('mjolnirgui/datasetdragdrop.list')
        stream = QtCore.QDataStream(encoded_data, QtCore.QIODevice.ReadOnly)

        new_items = []
        rows = 0
        while not stream.atEnd():
            item = QtCore.QByteArray()
            stream >> item
            item = int((bytes(item)).decode('utf-8)'))#pickle.loads(item)
            index = [ds.idx for ds in self.dataSets].index(item)
            item = self.dataSets[index]
            new_items.append((item, index))
            rows += 1

        self.lastDroppedItems = []
        for (text, index) in new_items:
            target_row = row
            if index < row:
                target_row += 1
            self.beginInsertRows(QtCore.QModelIndex(), target_row, target_row)
            self.dataSets.insert(target_row, text)
            self.endInsertRows()
            self.lastDroppedItems.append(text)
            row += 1
        self.pendingRemoveRowsAfterDrop = True
        return True

    def getCurrentDatasetIndex(self):
        indices = self.DataSet_DataSets_listView.selectedIndexes()
        
        if len(indices)==0:
            return None
        else:
            index = indices[0]
            if index.row()<self.rowCount(None):
                return index
            else:
                return None

    def getCurrentDatasetIndexRow(self):
        currentIndex = self.getCurrentDatasetIndex()
        if currentIndex is None:
            return None
        else:
            return currentIndex.row()

    def getCurrentDataSet(self):
        index = self.getCurrentDatasetIndex()
        return self.item(index)

    def selectLastDataSet(self):
        DataSets = self.rowCount(None)
        if DataSets!=0:
            index = self.index(self.rowCount(None)-1,0)
            self.DataSet_DataSets_listView.setCurrentIndex(index)


class SelectionModel(QtCore.QItemSelectionModel):
    """Selection model created from https://github.com/d1vanov/PyQt5-reorderable-list-model/blob/master/reorderable_list_model.py
    When drag/drop is used, selects newly inserted item"""
    def __init__(self, parent=None):
        QtCore.QItemSelectionModel.__init__(self, parent)

    def onModelItemsReordered(self):
        new_selection = QtCore.QItemSelection()
        new_index = QtCore.QModelIndex()
        for item in self.model().lastDroppedItems:
            row = self.model().rowForItem(item)
            if row is None:
                continue
            new_index = self.model().index(row, 0, QtCore.QModelIndex())
            new_selection.select(new_index, new_index)

        self.clearSelection()
        flags = QtCore.QItemSelectionModel.ClearAndSelect | \
                QtCore.QItemSelectionModel.Rows | \
                QtCore.QItemSelectionModel.Current
        self.select(new_selection, flags)
        self.setCurrentIndex(new_index, flags)


class Cut1DModel(QtCore.QAbstractListModel):
    def __init__(self, *args, dataCuts1D=None, Cut1D_listView=None, **kwargs):
        super(Cut1DModel, self).__init__(*args, **kwargs)
        self.dataCuts1D = dataCuts1D or []
        self.Cut1D_listView = Cut1D_listView
        
    def data(self, index, role):
        if role == Qt.DisplayRole or role == QtCore.Qt.EditRole:
            text = self.dataCuts1D[index.row()].name
            return text
        
    def getData(self,*args,**kwargs):
        return self.data(*args,**kwargs)

    def rowCount(self, index=None):
        return len(self.dataCuts1D)

    def append(self,Cut1D):
        self.dataCuts1D.append(Cut1D)
        self.selectLastCut1D()
        self.layoutChanged.emit()

    def delete(self,index):
        indices = [ind.row() for ind in index] # Extract numeric index, sort decending
        indices.sort(reverse=True)
        for ind in indices:
            try:
                del self.dataCuts1D[ind]
                self.layoutChanged.emit()
            except:
                pass

        QtWidgets.QApplication.processEvents()
        index = self.getCurrentCut1DIndex()
       
        if index is None:
            self.selectLastCut1D()
        else:
            if index.row()==self.rowCount(None):
                self.selectLastCut1D()

    def item(self,index):
        if not index is None:
            return self.dataCuts1D[index.row()]

    def setData(self, index, value, role=QtCore.Qt.EditRole):
        ds = self.item(index)
        if role == QtCore.Qt.EditRole:
            ds.name = value
            self.dataChanged.emit(index, index)
            return True
           
        return False

    def flags(self,index):
        return QtCore.Qt.ItemIsEditable | QtCore.Qt.ItemIsEnabled | QtCore.Qt.ItemIsSelectable

    def getCurrentCut1DIndex(self):
        indices = self.Cut1D_listView.selectedIndexes()
        
        if len(indices)==0:
            return None
        else:
            index = indices[0]
            if index.row()<self.rowCount(None):
                return index
            else:
                return None

    def getCurrentCut1DIndexRow(self):
        currentIndex = self.getCurrentCut1DIndex()
        if currentIndex is None:
            return None
        else:
            return currentIndex.row()

    def getCurrentCut1D(self):
        index = self.getCurrentCut1DIndex()
        return self.item(index)

    def selectLastCut1D(self):
        dataCuts1D = self.rowCount(None)
        if dataCuts1D!=0:
            index = self.index(self.rowCount(None)-1,0)
            self.Cut1D_listView.setCurrentIndex(index)


class BraggListModel(QtCore.QAbstractListModel):
    def __init__(self, *args, BraggList=None, braggList_listView=None, **kwargs):
        super(BraggListModel, self).__init__(*args, **kwargs)
        if BraggList is None:
            BraggList = []
        self._data = BraggList 
        self.braggList_listView = braggList_listView
        
    def data(self, index, role):
        if index.row()>=self.rowCount(): return 
        if index.row()<0: return 
        if index.column()>=1: return 
        if index.column()<-1: return 

        if role == Qt.DisplayRole or role == QtCore.Qt.EditRole:
            text = '\t'.join(map(str,self._data[index.row()]))
            return text
        
    def getData(self,*args,**kwargs):
        return self.data(*args,**kwargs)

    def getAllData(self):
        return self._data

    def rowCount(self, index=None):
        return len(self._data)

    def columnCount(self, index=None):
        return 0

    def append(self,BraggPoint):
        self._data.append(BraggPoint)
        self.selectLastBragg()
        self.layoutChanged.emit()

    def deleteAll(self):
        indices = [self.index(i,0) for i in range(self.rowCount())]
        self.delete(indices)

    def delete(self,index):
        indices = [ind.row() for ind in index] # Extract numeric index, sort decending
        indices.sort(reverse=True)
        for ind in indices:
            try:
                del self._data[ind]
                self.layoutChanged.emit()
            except:
                pass

        QtWidgets.QApplication.processEvents()
        #index = self.getCurrentBraggIndex()
       
        #if index is None:
        #    self.selectLastBragg()
        #else:
        #    if index.row()==self.rowCount(None):
        #        self.selectLastBragg()

    def item(self,index):
        if not index is None:
            return self._data[index.row()]

    #def setData(self, index, value, role=QtCore.Qt.EditRole):
    #    ds = self.item(index)
    #    if role == QtCore.Qt.EditRole:
    #        ds.name = value
    #        self.dataChanged.emit(index, index)
    #        return True
    #       
    #    return False

    def flags(self,index):
        return  QtCore.Qt.ItemIsSelectable | QtCore.Qt.ItemIsEnabled | QtCore.Qt.ItemIsEditable

    def getCurrentBraggIndex(self):
        indices = self.braggList_listView.selectedIndexes()
        
        if len(indices)==0:
            return None
        else:
            index = indices[0]
            if index.row()<self.rowCount(None):
                return index
            else:
                return None

    def getCurrentBraggIndexRow(self):
        currentIndex = self.getCurrentBraggIndex()
        if currentIndex is None:
            return None
        else:
            return currentIndex.row()

    def getCurrentBragg(self):
        index = self.getCurrentBraggIndex()
        return self.item(index)

    def selectLastBragg(self):
        bragg = self.rowCount(None)
        if bragg!=0:
            index = self.index(self.rowCount(None)-1,0)
            self.braggList_listView.setCurrentIndex(index)




class MaskModel(QtCore.QAbstractListModel):
    def __init__(self, *args, masks=None, Mask_listView=None, guiWindow=None, **kwargs):
        super(MaskModel, self).__init__(*args, **kwargs)
        self.masks = masks or []
        self.Mask_listView = Mask_listView
        self._combinedMask = None
        self.guiWindow = guiWindow

    @property
    def combinedMask(self):
        return self._combinedMask

    @combinedMask.getter
    def combinedMask(self):
        return self._combinedMask

    @combinedMask.setter
    def combinedMask(self,value):
        if not value is self._combinedMask:
            self._combinedMask = value
        
    def data(self, index, role):
        if role == Qt.DisplayRole or role == QtCore.Qt.EditRole:
            text = self.masks[index.row()].name
            return text
        
    def getData(self,*args,**kwargs):
        return self.data(*args,**kwargs)

    def rowCount(self, index=None):
        return len(self.masks)

    def generateValidName(self,ds):
        name = ds.name
        while name in self.getNames(): # name already exists.. This screws up drag/drop
            try:
                idx = int(name.split('_')[-1])
            except ValueError:
                name = name+'_1'
            else:
                name = name[:-(len(str(idx))+1)] + '_' + str(idx+1)
        return name

    def append(self,mask):
        mask.name = self.generateValidName(mask)

        if self.rowCount(None)>0:
            numbers = [d.idx for d in self.masks]
        else:
            numbers = [-1]
        mask.idx = np.max(numbers)+1
        self.masks.append(mask)
        self.selectLastMask()
        self.layoutChanged.emit()

    def delete(self,index):
        indices = [ind.row() for ind in index] # Extract numeric index, sort decending
        indices.sort(reverse=True)
        for ind in indices:
            try:
                del self.masks[ind]
                self.layoutChanged.emit()
            except:
                pass

        QtWidgets.QApplication.processEvents()
        selectIndex = self.getCurrentMaskIndex()
        if selectIndex is None:
            self.selectLastMask()
        else:
            if selectIndex.row()==self.rowCount(None):
                self.selectLastMask()

            else:
                idx = self.index(selectIndex.row(),0)# Hack to force index changed emit
                self.Mask_listView.setCurrentIndex(self.index(-1,0))
                self.Mask_listView.setCurrentIndex(idx)
        self.layoutChanged.emit()

    def item(self,index):   
        if not index is None:
            if not isinstance(index,list):
                try:
                    return self.masks[index.row()]
                except (AttributeError,IndexError):
                    return None
            else:
                index = np.asarray(index)
                if len(index) == 0:
                    return None
                indices = np.array([x.row() for x in index],dtype=int)
                return np.asarray(self.masks)[indices]

    def setData(self, index, value, role=QtCore.Qt.EditRole):
        mask = self.item(index)
        if role == QtCore.Qt.EditRole:
            mask.name = value
            self.dataChanged.emit(index, index)
            return True
           
        return False
    
    def getNames(self):
        return [m.name for m in self.masks]

    def flags(self,index):
        return QtCore.Qt.ItemIsEditable | QtCore.Qt.ItemIsEnabled | QtCore.Qt.ItemIsSelectable

    def getCurrentMaskIndex(self):
        indices = self.Mask_listView.selectedIndexes()
        
        if len(indices)==0:
            return None
        else:
            index = indices[0]
            if index.row()<self.rowCount(None):
                return index
            else:
                return None

    def getCurrentMaskIndexRow(self):
        currentIndex = self.getCurrentMaskIndex()
        if currentIndex is None:
            return None
        else:
            return currentIndex.row()

    def getMask(self):
        index = self.getCurrentMaskIndex()
        return self.item(index)

    def selectLastMask(self):
        
        if len(self.masks)>0:
            index = self.index(len(self.masks)-1,0)
        else:
            index = self.index(-1,0)
        self.Mask_listView.setCurrentIndex(index)
        



### Conversion from row call to attribute name

class defaultList(list):
    """Overloading of list object to always return list[0] if entry larger than lengths is called"""
    def __init__(self,*args,**kwargs):
        super(defaultList,self).__init__(*args,**kwargs)

    def __getitem__(self,index):
        if index>len(self)-1:
            return super(defaultList, self).__getitem__(0)
        else:
            return super(defaultList, self).__getitem__(index)

conversionList = defaultList(['name',
                            'ScanCommand'])


class defDict(dict):
    def __init__(self,*args,**kwargs):
        dict.__init__(self, args)
        #super(defDict,self).__init__(*args,**kwargs)


    def __getitem__(self, key):
        if not key in dict.keys(self):
            val = dict.__getitem__(self, 'default')
        else:
            val = dict.__getitem__(self, key)
        
        return val

    def __setitem__(self, key, val):
        
        dict.__setitem__(self, key, val)
    
    
class DummyIndexClass(object):
    def __init__(self,row):
        self._row = row

    def row(self):
        return self._row



class DataFileModel(QtCore.QAbstractListModel):

    dragDropFinished = QtCore.pyqtSignal()

    def __init__(self,DataSet_filenames_listView=None,dataSetModel=None,DataSet_DataSets_listView=None,guiWindow=None, **kwargs):
        super(DataFileModel, self).__init__(**kwargs)
        self.dataSetModel = dataSetModel
        self.DataSet_DataSets_listView = DataSet_DataSets_listView
        self.DataSet_filenames_listView = DataSet_filenames_listView
        self.guiWindow = guiWindow
        self.lastDroppedItems = []
        self.pendingRemoveRowsAfterDrop = False

        self.IconDict = defDict()
        self.IconDict['default']=QtGui.QImage(self.guiWindow.AppContext.get_resource('Icons/Own/document.png'))
        self.IconDict['hdf']=QtGui.QImage(self.guiWindow.AppContext.get_resource('Icons/Own/HDF_logo_16.png'))
        self.IconDict['nxs']=QtGui.QImage(self.guiWindow.AppContext.get_resource('Icons/Own/NXS_logo_16.png'))

    def rowForItem(self, item):
        '''
        rowForItem method returns the row corresponding to the passed in item
        or None if no such item exists in the model
        '''
        try:
            row = list(self.dataSetModel.item(self.getCurrentDatasetIndex())).index(item)
        except ValueError:
            return None
        return row
        
    def data(self, index, role=Qt.ItemDataRole):
        try:
            ds = self.dataSetModel.item(self.getCurrentDatasetIndex())
        except AttributeError:
            return None
        
        ds = self.dataSetModel.item(self.getCurrentDatasetIndex())
        row = index.row()
        if row>= len(ds):
            return None

        if role == Qt.DisplayRole or role == Qt.EditRole:
            
            text = ds[row].name
            return text
        
        if role == Qt.DecorationRole:
            t = ds[row].type
            return self.IconDict[t]

        if role == Qt.ItemDataRole:
            return ds[row]

    def insertRows(self, row, count, index):
        if index.isValid():
            return False
        if count <= 0:
            return False
        # inserting 'count' empty rows starting at 'row'
        self.beginInsertRows(QtCore.QModelIndex(), row, row + count - 1)
        for i in range(0, count):
            self.dataSetModel.item(self.getCurrentDatasetIndex()).insert(row + i, None)
        self.endInsertRows()
        return True

    def removeRows(self, row, count, index):
        if index.isValid():
            return False
        if count <= 0:
            return False
        # num_rows = self.rowCount(QtCore.QModelIndex())
        self.beginRemoveRows(QtCore.QModelIndex(), row, row + count - 1)
        for i in range(count, 0, -1):
             self.dataSetModel.item(self.getCurrentDatasetIndex()).pop(row - i + 1)
        self.endRemoveRows()

        if self.pendingRemoveRowsAfterDrop:
            '''
            If we got here, it means this call to removeRows is the automatic
            'cleanup' action after drag-n-drop performed by Qt
            '''
            self.pendingRemoveRowsAfterDrop = False
            self.dragDropFinished.emit()

        return True

    def index(self, row, column, parent=None):
        if row < 0 or row >= self.rowCount(None):
            return QtCore.QModelIndex()
        return self.createIndex(row, column)

    def parent(self, index):
        return QtCore.QModelIndex()

    def supportedDropActions(self):
        return QtCore.Qt.MoveAction 

    def getCurrentDatasetIndex(self):
        if hasattr(self.DataSet_DataSets_listView,'selectedIndexes'): # If DataSet_DataSets_listView is listView
            indices = self.DataSet_DataSets_listView.selectedIndexes()
            
            if indices is None:
                return None
            else:
                if len(indices)==0:
                    return None
                else:
                    return indices[0]
        else: # If it is a combo box, needed for subrtaction manager. Mimic a proper index
            return DummyIndexClass(self.DataSet_DataSets_listView.currentIndex())

    def getCurrentDatasetIndexRow(self):
        currentIndex = self.getCurrentDatasetIndex()
        if currentIndex is None:
            return None
        else:
            return currentIndex.row()

    def getCurrentDatafileIndexs(self):
        indices = self.DataSet_filenames_listView.selectedIndexes()
        
        if indices is None:
            return None
        else:
            idxs = self.DataSet_filenames_listView.selectedIndexes()
            
            if np.all([idx.row()<self.rowCount(None) for idx in idxs]):
                return idxs
            else:
                return None

    def getCurrentDatafileIndexRows(self):
        currentIndeces = self.getCurrentDatafileIndexs()
        if currentIndeces is None:
            return None
        else:
            return [idx.row() for idx in currentIndeces]

    def getCurrentDatafiles(self,raw=False):
        """return current selected data files. If raw is true, return only raw."""
        indexRows = self.getCurrentDatafileIndexRows()
        if indexRows is None:
            return None
        if len(indexRows)==0:
            return None
        if np.any([idx is None for idx in indexRows]):
            return None
        else:
            ds = self.dataSetModel.item(self.getCurrentDatasetIndex())
            if raw:
                dfs = [ds[idx] if not ds[idx].type == 'nxs' else ds[idx].original_file for idx in indexRows]
            else:
                dfs = [ds[idx] for idx in indexRows]
            return dfs

    def getData(self,raw = False):
        """Return all data files in dataset independent of selection. If raw is true, return only raw."""
        ds = self.dataSetModel.item(self.getCurrentDatasetIndex())
        if ds is None:
            return []
        if len(ds) == 0:
            return []
        if raw:
            dfs = [df if not df.type == 'nxs' else df.original_file for df in ds]
        else:
            dfs = [df for df in ds]
        
        return dfs

    def getCurrentDataSet(self):
        return self.dataSetModel.item(self.getCurrentDatasetIndex())

    def rowCount(self, index):
        try:
            ds = self.dataSetModel.item(self.getCurrentDatasetIndex())
        except IndexError:
            return 0
        if ds is None:
            return 0
        else:
            return len(ds)
        
    def selectFirstDataFile(self):
        Files = self.rowCount(None)
        if Files!=0:
            index = self.index(0,0)
            self.DataSet_filenames_listView.setCurrentIndex(index)
        else:
            self.DataSet_filenames_listView.clearSelection()

    def updateCurrentDataSetIndex(self):
        self.selectFirstDataFile()
        self.layoutChanged.emit()

    def setData(self, index, value, role=QtCore.Qt.EditRole):
        ds = self.dataSetModel.item(self.getCurrentDatasetIndex())
        df = ds[index.row()]
        
        if role == QtCore.Qt.EditRole:
            df.name = value
            
            return True
            
        return False


    def flags(self,index):
        if not index.isValid():
            return QtCore.Qt.ItemIsEnabled | QtCore.Qt.ItemIsDropEnabled
        return QtCore.Qt.ItemIsEditable | QtCore.Qt.ItemIsEnabled | QtCore.Qt.ItemIsSelectable | \
            QtCore.Qt.ItemIsDragEnabled | QtCore.Qt.ItemIsDropEnabled
        #return QtCore.Qt.ItemIsEditable | QtCore.Qt.ItemIsEnabled | QtCore.Qt.ItemIsSelectable


    def delete(self,idx=None):
        if idx is None:
            ds = self.dataSetModel.item(self.getCurrentDatasetIndex())
        else:
            ds = self.dataSetModel.item(idx)
        indices = np.array(self.getCurrentDatafileIndexRows())
        if np.all([row<len(ds) for row in indices]):
            for idx in indices:
                df = ds[idx]
                if df.type == 'nxs': #i.e. converted
                    del ds._dataFiles[idx]
                del ds[idx]
                indices-=1
            if not self.pendingRemoveRowsAfterDrop:
                self.layoutChanged.emit()
            self.guiWindow.updateDataFileLabels()


    def add(self,files,guiWindow=None):
        ds = self.dataSetModel.item(self.getCurrentDatasetIndex())
        dfs = []
        binning = None
        if len(ds)>0: # non-empty dataset
            if ds[0].type=='nxs': # Converted dataset
                binning = ds[0].binning

        if not guiWindow is None:
            if binning is None: # length is simply number of files
                length = len(files)
            else:
                length = 2*len(files) # both raw and converted
            guiWindow.setProgressBarMaximum(length)
            printFunction = guiWindow.writeToStatus
        else:
            printFunction = None
        
        for i,f in enumerate(files):
            newFile = GuiDataFile(f)
            dfs.append(newFile)
            if not guiWindow is None:
                guiWindow.setProgressBarValue(i+1)
            if not binning is None:
                convFile = newFile.convert(binning=binning,printFunction=printFunction)
                dfs.append(convFile)
                if not guiWindow is None:
                    guiWindow.setProgressBarValue(i+1)
            

        ds.append(dfs)
        if not self.pendingRemoveRowsAfterDrop:
            self.layoutChanged.emit()
        guiWindow.updateDataFileLabels()


    def mimeTypes(self):
        return ['mjolnirgui/datafiledragdrop.list']

    def mimeData(self, indexes):
        mimedata = QtCore.QMimeData()
        encoded_data = QtCore.QByteArray()
        stream = QtCore.QDataStream(encoded_data, QtCore.QIODevice.WriteOnly)
        items = []
        for index in indexes:
            if index.isValid():
                item = self.data(index, QtCore.Qt.ItemDataRole)
                items.append(item)
                
        for item in items:
            stream << QtCore.QByteArray(str(item.idx).encode('utf-8'))#pickle.dumps(item))
        mimedata.setData('mjolnirgui/datafiledragdrop.list', encoded_data)
        return mimedata

    def dropMimeData(self, data, action, row, column, parent):
        if action == QtCore.Qt.IgnoreAction:
            return True
        if not data.hasFormat('mjolnirgui/datafiledragdrop.list'):
            return False
        if column > 0:
            return False

        num_rows = self.rowCount(QtCore.QModelIndex())
        if num_rows <= 0:
            return False

        if row < 0:
            if parent.isValid():
                row = parent.row()
            else:
                return False

        encoded_data = data.data('mjolnirgui/datafiledragdrop.list')
        stream = QtCore.QDataStream(encoded_data, QtCore.QIODevice.ReadOnly)

        new_items = []
        rows = 0
        while not stream.atEnd():
            item = QtCore.QByteArray()
            stream >> item
            item = int(bytes(item).decode('utf-8'))#pickle.loads(item)
            index = [df.idx for df in self.dataSetModel.item(self.getCurrentDatasetIndex())].index(item)

            item = self.dataSetModel.item(self.getCurrentDatasetIndex())[index]
            new_items.append((item, index))
            rows += 1

        self.lastDroppedItems = []
        for (text, index) in new_items:
            target_row = row
            if index < row:
                target_row += 1
            self.beginInsertRows(QtCore.QModelIndex(), target_row, target_row)
            self.dataSetModel.item(self.getCurrentDatasetIndex()).insert(target_row, text)
            self.endInsertRows()
            self.lastDroppedItems.append(text)
            row += 1
        self.pendingRemoveRowsAfterDrop = True
        return True


def getAttribute(obj,location):
    if not '/' in location:
        return getattr(obj,location)
    
    splitLocation = location.split('/')
    newLocation = '/'.join(splitLocation[1:])
    newObj = getattr(obj,splitLocation[0])
    return getAttribute(newObj,newLocation)


def formatValueArray(array, formatString = '{:.2f} [{:.2f}  -  {:.2f}]'):
    """Format array to string using mean, min, max"""
    try:
        array = np.concatenate(array)
    except ValueError:
        return 'N/A'
    mean = np.mean(array)
    max_ = np.max(array)
    min_ = np.min(array)

    return formatString.format(mean,min_,max_)

def formatValueArrayDirection(array, formatString = '{:.2f} [{:.2f}  -  {:.2f}]'):
    """Format array to string using mean, min, max as well as direction (for A3)"""
    try:
        array = np.concatenate(array)
    except ValueError:
        return 'N/A'
    mean = np.mean(array)
    max_ = np.max(array)
    min_ = np.min(array)
    diff = np.diff(array.flatten())
    if np.all(diff>0):
        direction = ' Scan dir: +'
    elif np.all(diff<0):
        direction = ' Scan dir: -'
    else:
        direction = ''


    return formatString.format(mean,min_,max_)+direction

def formatTextArray(array):
    """Return string only if all entries are equal"""
    if len(array)==1:
        return array[0].strip()
        
    if np.all([array[0]==a for a in array]):
        return array[0].strip()
    else:
        return 'Multiple Values'

def formatTextArrayAdder(array):
    """Return list of all values"""
    array = np.array(array).flatten()
    if len(array)==1:
        return array[0]
    
    Set = set(array) # Set only conains one of each
    return ', '.join([str(s) for s in Set])


def formatVector(array):
    a = np.array(array,dtype=float)
    if len(a.shape)==1: # only one input
        return M_tools.generateLabel(a)
    else:
        if a.shape[1]>1: # more than one array
            if np.all([np.equal(a[0],x) for x in a[1:]]): # all equal to a[0]
                return M_tools.generateLabel(a[0])
            else:
                return 'N/A'
        else:
            return M_tools.generateLabel(a)


def formatRaw(array):
    if len(array) == 1:
        return str(array[0])
    return str(array)


Info = namedtuple('Info','location baseText formatter')
name = Info('sample/name','Sample: ',formatTextArray)
title = Info('title','Title: ',formatTextArray)
projectionVector1 = Info('sample/projectionVector1','Projection 1: ',formatVector)
projectionVector2 = Info('sample/projectionVector2','Projection 2: ',formatVector)
A3 = Info('A3','A3 [deg]: ',formatValueArrayDirection)
tt = Info('twotheta','2θ [deg]: ',formatValueArray)
magneticField = Info('magneticField','Magnetic Field [B]: ',formatValueArray)
temperature = Info('temperature','Temperature [K]: ',formatValueArray)
scanCommand = Info('scanCommand','Command: ',formatTextArray)
scanSteps = Info('scanSteps','Scan steps: ',formatRaw)
scanParameters = Info('scanParameters','Parameter: ',formatTextArrayAdder)
comment = Info('comment','Comment: ',formatTextArray)
binning = Info('binning','Binning: ',formatRaw)
Ei = Info('Ei','Ei [meV]: ',formatValueArray)
countingTime = Info('Time', 'Step time [s]: ',formatValueArray)
startTime = Info('startTime', 'Start time: ', formatTextArrayAdder)
endTime = Info('endTime', 'End time: ', formatTextArrayAdder)

settings = {'sample/name':name,'title':title,'sample/projectionVector1':projectionVector1,'sample/projectionVector2':projectionVector2,'Ei':Ei, 'A3':A3,'twotheta':tt, 'magneticField':magneticField,'temperature':temperature,
            'comment':comment, 'binning':binning,'scanCommand':scanCommand, 'scanSteps':scanSteps, 'scanParameters':scanParameters,
            'Time':countingTime,'startTime':startTime, 'endTime':endTime}

subtractionSettings = {'sample/name':name,'title':title,'sample/projectionVector1':projectionVector1,'sample/projectionVector2':projectionVector2,'Ei':Ei, 'A3':A3,'twotheta':tt, 'magneticField':magneticField,'temperature':temperature, 'scanCommand':scanCommand, 'scanSteps':scanSteps, 'scanParameters':scanParameters}

class DataFileInfoModel(QtCore.QAbstractListModel):
    def __init__(self, *args, DataSet_filenames_listView=None,dataSetModel=None,DataSet_DataSets_listView=None,dataFileModel=None,guiWindow=None, **kwargs):
        super(DataFileInfoModel, self).__init__(*args, **kwargs)
        self.dataSetModel = dataSetModel
        self.dataFileModel = dataFileModel
        self.DataSet_DataSets_listView = DataSet_DataSets_listView
        self.DataSet_filenames_listView = DataSet_filenames_listView
        self.guiWindow = guiWindow
        self.possibleSettings = settings
        self.infos = []

        
    def data(self, index, role):
        if role == Qt.DisplayRole or role == Qt.EditRole:
            dfs = self.dataFileModel.getCurrentDatafiles()
            I = index.row()
            info = self.infos[I]
            if not dfs is None:
                
                if len(dfs)==1:
                    data = [getAttribute(dfs[0],info.location)]
                else:
                    data =[getAttribute(df,info.location) for df in dfs]
                return info.baseText+info.formatter(data)
            else:
                return info.baseText
        elif role == Qt.ItemDataRole:
            I = index.row()
            info = self.infos[I]
            dfs = self.dataFileModel.getCurrentDatafiles()
            if not dfs is None:
                df = dfs[0]
            else:
                return True
            
            if hasattr(df,info.location+'_check'):
                return getattr(df,info.location+'_check')
            else:
                return True


        
        
    def rowCount(self,index):
        return len(self.infos)

    @property
    def infos(self):
        return self._infos

    @infos.setter
    def infos(self, newSettings):
        newSettings = np.array(newSettings)
        inside = np.array([I in self.possibleSettings.keys() for I in newSettings],dtype=bool)
        if not np.all(inside):
            outside = np.array(1-inside,dtype=bool)
            warnings.warn('Wanted setting(s) {} not found. Allowed are {}'.format(newSettings[outside],settings.keys()))
        self._infos = [settings[I] for I in newSettings[inside]]


    def possibleInfos(self):
        return [key for key in self.possibleSettings]


    def currentInfos(self):
        return [setting.location for setting in self.infos]
        
    def settingsDialog(self):
        return self.possibleSettings,self.infos

ValidColour =  QtGui.QColor(255, 255, 255, 0)
InvalidColour = QtGui.QColor(255, 0, 0, 120)
class MatplotlibFigureListDelegate(QtWidgets.QStyledItemDelegate):
    def __init__(self, parent=None, *args):
        """Delegate colouring the background of data files with the BG colour given by df.BGColor, default is ValidColor"""
        QtWidgets.QStyledItemDelegate.__init__(self, parent, *args)


    def paint(self, painter, option, index):
        super(MatplotlibFigureListDelegate,self).paint(painter, option, index)
        if True:
            
            figure = index.model().data(index,Qt.ItemDataRole)
            #print(figure._title,option._state)
            painter.save()
            painter.setPen(QtGui.QPen(Qt.NoPen))
            if not hasattr(figure,'closed'):
                c = ValidColour
            elif figure.closed: # If figure is closed, color the backgroud red
                c = InvalidColour
            else:
                c = ValidColour
            painter.setBrush(QtGui.QBrush(c))

            # Draw the background rectangle            
            painter.drawRect(option.rect)
            painter.restore()

class MatplotlibFigureList(QtCore.QAbstractListModel):
    def __init__(self, *args, combobox=None, **kwargs):
        super(MatplotlibFigureList, self).__init__(*args, **kwargs)
        self.figures = []
        
        self.view = combobox
        
    
    def setup(self):
        self.view.setModel(self)

    def data(self, index, role):
        if role == Qt.DisplayRole or role == QtCore.Qt.EditRole:
            if hasattr(self.figures[index.row()],'_title'):
                text = self.figures[index.row()]._title
            else:
                text = 'NoTitle...'
            return text

        if role == Qt.ItemDataRole:
            return self.figures[index.row()]

    def getData(self,*args,**kwargs):
        return self.data(*args,**kwargs)

    def rowCount(self, index=None):
        return len(self.figures)
        

    def append(self,figure):
        if not hasattr(figure,'closed'):
            figure.closed = False
        self.figures.append(figure)

        self.selectLastFigure()
        self.layoutChanged.emit()

    
    #def delete(self,index):
    #    indices = [ind.row() for ind in index] # Extract numeric index, sort decending
    #    indices.sort(reverse=True)
    #    for ind in indices:
    #        try:
    #            del self.figures[ind]
    #            self.layoutChanged.emit()
    #        except:
    #            pass

    #    QtWidgets.QApplication.processEvents()
    #    index = self.getCurrentFigureIndex()
    #   
    #    if index is None:
    #        self.selectLastCut1D()
    #    else:
    #        if index.row()==self.rowCount(None):
    #            self.selectLastFigure()

    #def item(self,index,figureType=None):
    #    if not index is None:
    #        return self.figure[index.row()]

    def flags(self,index):
        #figure = self.figures[index.row()]
        #if figure.closed:
        #    return QtCore.Qt.ItemIsSelectable
        return QtCore.Qt.ItemIsSelectable | QtCore.Qt.ItemIsEnabled #QtCore.Qt.ItemIsEditable  | 

    def getCurrentFigureIndex(self):
        index = self.view.currentIndex()
        if index is None:
            return None
        else:
            if index<self.rowCount(None):
                return index
            else:
                return None
        
    def getCurrentFigure(self):
        index = self.view.currentIndex()
        if index is None:
            return None
        else:
            if index<self.rowCount(None) and index>-1:
                return self.figures[index]
            else:
                return None

    #def getCurrentCut1DIndexRow(self):
    #    currentIndex = self.getCurrentCut1DIndex()
    #    if currentIndex is None:
    #        return None
    #    else:
    #        return currentIndex.row()

    #def getCurrentCut1D(self):
    #    index = self.getCurrentCut1DIndex()
    #    return self.item(index)

    def selectLastFigure(self):
        figures = self.rowCount(None)
        if figures!=0:
            index = self.index(self.rowCount(None)-1,0)
            self.view.setCurrentIndex(index.row())

    def closeAll(self):
        for fig in self.figures:
            self.close(fig)

    def close(self,figure):
        if hasattr(figure,'ax'):
            plt.close(figure.ax.get_figure())
            figure.closed = True
        else:
            try:
                plt.close(figure)
                figure.closed = True
            except:
                figure.parent().close()
                figure.closed = True
        self.layoutChanged.emit()
        #del figure


import sys
sys.path.append('..')

from os import path


from MJOLNIR._tools import DSpacing,ScatteringAngle
try:
    from _tools import loadUI
except ModuleNotFoundError:
    from MJOLNIRGui.src.main.python._tools import loadUI
# Handles all functionality related to the CalculatorSimpleManager. 


CalculatorSimpleManagerBase, CalculatorSimpleManagerForm = loadUI('Calculator_Simple.ui')

# All of this connects the buttons and their functions to the main window.
       

def textChangedA4(manager):
    A4 = manager.a4_spinBox.value()
    E = manager.e_spinBox.value()
    manager.d_spinBox.setValue(DSpacing(A4,Energy=E))
    

def textChangedEnergy(manager):
    E = manager.e_spinBox.value()
    D = manager.d_spinBox.value()
    manager.a4_spinBox.setValue(ScatteringAngle(D,Energy=E))
    

def textChangedD(manager):
    D = manager.d_spinBox.value()
    E = manager.e_spinBox.value()
    manager.a4_spinBox.setValue(ScatteringAngle(D,Energy=E))


def onFocus(self,event,others):
    for o in others:
        try:
            o.valueChanged.disconnect()
        except TypeError: # If no connection to remove
            pass
    self.valueChanged.connect(self.onChangeFunction)
    self.old_focusInEvent(event)




class CalculatorSimpleManager(CalculatorSimpleManagerBase, CalculatorSimpleManagerForm):
    def __init__(self, parent=None, guiWindow=None):
        super(CalculatorSimpleManager, self).__init__(parent)
        self.setupUi(self)
        self.guiWindow = guiWindow

        self.initCalculatorSimpleManager()
        
    def initCalculatorSimpleManager(self):    
        self.setup()
        

    def setup(self):
        # Add updating functions to be called when text is changed
        self.a4_spinBox.onChangeFunction = lambda: textChangedA4(self)
        self.e_spinBox.onChangeFunction = lambda: textChangedEnergy(self)
        self.d_spinBox.onChangeFunction = lambda: textChangedD(self)
        
        # Move default focusInEvent
        self.a4_spinBox.old_focusInEvent = self.a4_spinBox.focusInEvent
        self.e_spinBox.old_focusInEvent = self.e_spinBox.focusInEvent
        self.d_spinBox.old_focusInEvent = self.d_spinBox.focusInEvent

        # Update to new 
        self.a4_spinBox.focusInEvent= lambda event: onFocus(self.a4_spinBox,event,[self.e_spinBox,self.d_spinBox])
        self.e_spinBox.focusInEvent = lambda event: onFocus(self.e_spinBox,event,[self.a4_spinBox,self.d_spinBox])
        self.d_spinBox.focusInEvent = lambda event: onFocus(self.d_spinBox,event,[self.a4_spinBox,self.e_spinBox])
        
